# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPDInstance, \
                                                          OpenBGPDPreviousInstance, \
                                                          OpenBGPDLatestInstance

class RichConfigExampleScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    MOCK_RIPE_RPKI_CACHE = False
    MOCK_PEERING_DB = False
    MOCK_ARIN_DB_DUMP = False
    MOCK_REGISTROBR_DB_DUMP = False

    AS_SET = {
        "AS3333": [3333],
        "AS10745": [10745],
    }
    R_SET = {
        "AS10745": [
            "AS10745_allowed_prefixes"
        ],
        "AS3333": [
            "AS3333_allowed_prefixes"
        ],
    }
    RTT = {
        "192.0.2.11": 114,
        "192.0.2.22": 224,
        "2001:db8:1:1::22": 226
    }

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance()
        ]

    def set_instance_variables(self):
        self.rs = self._get_instance_by_name("rs")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

class RichConfigExampleScenarioBIRD(RichConfigExampleScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder
    TARGET_VERSION = None
    IP_VER = None

    @classmethod
    def _get_local_file(cls):
        return "bird.client.local"

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER,
                                     local_files=["client"],
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bird/bird.conf"
                ),
                (
                    cls.use_static_file(cls._get_local_file()),
                    "/etc/bird/client.local"
                )
            ]
        )

class RichConfigExampleScenarioBIRD2(RichConfigExampleScenarioBIRD):
    __test__ = False

    TARGET_VERSION = "2.0.7"

    @classmethod
    def _get_local_file(cls):
        return "bird2.client.local"

class RichConfigExampleScenarioOpenBGPD(LiveScenario_TagRejectPolicy,
                                        RichConfigExampleScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder

    TARGET_VERSION = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bgpd.conf"
                )
            ]
        )

class RichConfigExampleScenarioOpenBGPDPrevious(RichConfigExampleScenarioOpenBGPD):

    __test__ = False

    TARGET_VERSION = OpenBGPDPreviousInstance.BGP_SPEAKER_VERSION

class RichConfigExampleScenarioOpenBGPDLatest(RichConfigExampleScenarioOpenBGPD):

    __test__ = False

    TARGET_VERSION = OpenBGPDLatestInstance.BGP_SPEAKER_VERSION



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

try:
    import mock
except ImportError:
    import unittest.mock as mock
import os
import unittest

import yaml

from .cfg_base import TestConfigParserBase
from pierky.arouteserver.config.general import ConfigParserGeneral
from pierky.arouteserver.errors import ConfigError


class TestConfigParserGeneral(TestConfigParserBase):

    FILE_PATH = "config.d/general.yml"
    CONFIG_PARSER_CLASS = ConfigParserGeneral
    SHORT_DESCR = "General config parser"

    VALID_STD_COMMS = ("65534:666", "1:1", "0:0", "1:0", "65534:65535", "rs_as:1")
    VALID_LRG_COMMS = ("65535:666:0", "1:1:1", "0:0:0", "1:0:65535", "4294967295:4294967295:4294967295", "rs_as:1:2")
    VALID_EXT_COMMS = ("rt:1:0", "rt:65535:666", "ro:65535:666", "rt:rs_as:3")

    def test_valid_cfg(self):
        """{}: valid configuration"""
        self._contains_err()

    def test_unknown_statement(self):
        """{}: unknown statement"""
        self.cfg["test"] = "test"
        self._contains_err("Unknown statement at 'cfg' level: 'test'.")

    def test_rs_as(self):
        """{}: rs_as"""
        self.assertEqual(self.cfg["rs_as"], 999)
        for asn in (-1, 0, "test"):
            self.cfg["rs_as"] = asn
            self._contains_err("Error parsing 'rs_as' at 'cfg' level - Invalid ASN: {}".format(str(asn)))

        for asn in (1, 65535, 4294967295):
            self.cfg["rs_as"] = asn
            self._contains_err()

        self._test_mandatory(self.cfg, "rs_as")

    def test_router_id(self):
        """{}: router_id"""
        self.assertEqual(self.cfg["router_id"], "192.0.2.2")
        for v in ("1.0.0.1", "10.0.0.1"):
            self.cfg["router_id"] = v
            self._contains_err()
        for v in ("10.0.0.1/24", "fe80::1", "2001:db8::1", "test"):
            self.cfg["router_id"] = v
            self._contains_err("Error parsing 'router_id' at 'cfg' level - Invalid IPv4 address: {}.".format(v))
        self._test_mandatory(self.cfg, "router_id")

    def test_prepend_rs_as(self):
        """{}: prepend_rs_as"""
        self.assertEqual(self.cfg["prepend_rs_as"], False)
        self._test_bool_val(self.cfg, "prepend_rs_as")
        self._test_mandatory(self.cfg, "prepend_rs_as", has_default=True)

    def test_path_hiding(self):
        """{}: path_hiding"""
        self._test_bool_val(self.cfg, "path_hiding")
        self._test_optional(self.cfg, "path_hiding")

    def test_passive(self):
        """{}: passive"""
        self.assertEqual(self.cfg["passive"], True)
        self._test_bool_val(self.cfg, "passive")
        self._test_mandatory(self.cfg, "passive", has_default=True)

    def test_gtsm(self):
        """{}: gtsm"""
        self.assertEqual(self.cfg["gtsm"], False)
        self._test_bool_val(self.cfg, "gtsm")
        self._test_mandatory(self.cfg, "gtsm", has_default=True)

    def test_add_path(self):
        """{}: add_path"""
        self.assertEqual(self.cfg["add_path"], False)
        self._test_bool_val(self.cfg, "add_path")
        self._test_mandatory(self.cfg, "add_path", has_default=True)

    def test_next_hop_policy(self):
        """{}: next_hop.policy"""
        self.assertEqual(self.cfg["filtering"]["next_hop"]["policy"], "strict")
        self._test_option(self.cfg["filtering"]["next_hop"], "policy", ("strict", "same-as"))
        self._test_mandatory(self.cfg["filtering"]["next_hop"], "policy", has_default=True)

    def test_next_hop_policy_pre_0_6_0(self):
        """{}: next_hop_policy (pre v0.6.0 format)"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  filtering:\n"
            "    next_hop_policy: 'same-as'\n"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["filtering"]["next_hop"]["policy"], "same-as")

    def test_ipv4_pref_len(self):
        """{}: ipv4_pref_len"""
        self.assertEqual(self.cfg["filtering"]["ipv4_pref_len"]["min"], 8)
        self.assertEqual(self.cfg["filtering"]["ipv4_pref_len"]["max"], 24)
        self.cfg["filtering"]["ipv4_pref_len"]["max"] = 32
        self._test_ip_min_max_len(self.cfg["filtering"], "ipv4_pref_len", "min", 4, (0, 1, 32), (-1, 33, "a"))

        self.cfg["filtering"]["ipv4_pref_len"]["min"] = 0
        self._test_ip_min_max_len(self.cfg["filtering"], "ipv4_pref_len", "max", 4, (0, 1, 32), (-1, 33, "a"))

        for pair in ((0, 32), (31, 32), (32, 32)):
            self.cfg["filtering"]["ipv4_pref_len"]["min"] = pair[0]
            self.cfg["filtering"]["ipv4_pref_len"]["max"] = pair[1]
            self._contains_err()
        for pair in ((10, 5), (32, 0)):
            self.cfg["filtering"]["ipv4_pref_len"]["min"] = pair[0]
            self.cfg["filtering"]["ipv4_pref_len"]["max"] = pair[1]
            self._contains_err("Error parsing 'ipv4_pref_len' at 'cfg.filtering' level - In the IPv4 min/max length, the value of 'min' must be <= the value of 'max'.")

    def test_ipv6_pref_len(self):
        """{}: ipv6_pref_len"""
        self.assertEqual(self.cfg["filtering"]["ipv6_pref_len"]["min"], 12)
        self.assertEqual(self.cfg["filtering"]["ipv6_pref_len"]["max"], 48)
        self.cfg["filtering"]["ipv6_pref_len"]["max"] = 128
        self._test_ip_min_max_len(self.cfg["filtering"], "ipv6_pref_len", "min", 6, (0, 1, 32, 64, 128), (-1, 129, "a"))

        self.cfg["filtering"]["ipv6_pref_len"]["min"] = 0
        self._test_ip_min_max_len(self.cfg["filtering"], "ipv6_pref_len", "max", 6, (0, 1, 32, 64, 128), (-1, 129, "a"))

        for pair in ((0, 64), (32, 48), (32, 32), (128, 128)):
            self.cfg["filtering"]["ipv6_pref_len"]["min"] = pair[0]
            self.cfg["filtering"]["ipv6_pref_len"]["max"] = pair[1]
            self._contains_err()
        for pair in ((10, 5), (32, 0), (128, 0), (128, 64)):
            self.cfg["filtering"]["ipv6_pref_len"]["min"] = pair[0]
            self.cfg["filtering"]["ipv6_pref_len"]["max"] = pair[1]
            self._contains_err("Error parsing 'ipv6_pref_len' at 'cfg.filtering' level - In the IPv6 min/max length, the value of 'min' must be <= the value of 'max'.")

    def test_global_black_list_pref(self):
        """{}: global_black_list_pref"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  filtering:\n"
            "    global_black_list_pref:\n"
            "      - prefix: 192.0.2.0\n"
            "        length: 24\n"
            "        comment: 'Local network'\n"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["filtering"]["global_black_list_pref"][0],
                         {
                             "prefix": "192.0.2.0",
                             "length": 24,
                             "comment": "Local network",
                             "exact": False,
                             "ge": None,
                             "le": None,
                             "max_length": 32
                         })
        self._test_optional(self.cfg["filtering"], "global_black_list_pref")

    def test_max_as_path_len(self):
        """{}: max_as_path_len"""
        self.assertEqual(self.cfg["filtering"]["max_as_path_len"], 32)
        for v in (0, 65):
            self.cfg["filtering"]["max_as_path_len"] = v
            self._contains_err("Error parsing 'max_as_path_len' at 'cfg.filtering' level - Invalid max_as_path_len: must be an integer between 1 and 64.")
        for v in (1, 64, 32):
            self.cfg["filtering"]["max_as_path_len"] = v
            self._contains_err()
        self._test_mandatory(self.cfg["filtering"], "max_as_path_len", has_default=True)

    def test_reject_invalid_as_in_as_path(self):
        """{}: reject_invalid_as_in_as_path"""
        self.assertEqual(self.cfg["filtering"]["reject_invalid_as_in_as_path"], True)
        self._test_bool_val(self.cfg["filtering"], "reject_invalid_as_in_as_path")
        self._test_mandatory(self.cfg["filtering"], "reject_invalid_as_in_as_path", has_default=True)

    def test_tag_as_set(self):
        """{}: tag_as_set"""
        self.assertEqual(self.cfg["filtering"]["irrdb"]["tag_as_set"], True)
        self._test_bool_val(self.cfg["filtering"]["irrdb"], "tag_as_set")
        self._test_mandatory(self.cfg["filtering"]["irrdb"], "tag_as_set", has_default=True)

    def test_enforce_origin_in_as_set(self):
        """{}: enforce_origin_in_as_set"""
        self.assertEqual(self.cfg["filtering"]["irrdb"]["enforce_origin_in_as_set"], True)
        self._test_bool_val(self.cfg["filtering"]["irrdb"], "enforce_origin_in_as_set")
        self._test_mandatory(self.cfg["filtering"]["irrdb"], "enforce_origin_in_as_set", has_default=True)

    def test_enforce_prefix_in_as_set(self):
        """{}: enforce_prefix_in_as_set"""
        self.assertEqual(self.cfg["filtering"]["irrdb"]["enforce_prefix_in_as_set"], True)
        self._test_bool_val(self.cfg["filtering"]["irrdb"], "enforce_prefix_in_as_set")
        self._test_mandatory(self.cfg["filtering"]["irrdb"], "enforce_prefix_in_as_set", has_default=True)

    def test_allow_longer_prefixes(self):
        """{}: allow_longer_prefixes"""
        self.assertEqual(self.cfg["filtering"]["irrdb"]["allow_longer_prefixes"], False)
        self._test_bool_val(self.cfg["filtering"]["irrdb"], "allow_longer_prefixes")
        self._test_mandatory(self.cfg["filtering"]["irrdb"], "allow_longer_prefixes", has_default=True)

    def test_use_rpki_roas_as_route_objects_enabled(self):
        """{}: use_rpki_roas_as_route_objects.enabled"""
        self.assertEqual(self.cfg["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["enabled"], False)
        self._test_bool_val(self.cfg["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"], "enabled")
        self._test_mandatory(self.cfg["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"], "enabled", has_default=True)

    def test_use_rpki_roas_source(self):
        """{}: rpki_roas.source"""
        self.assertEqual(self.cfg["rpki_roas"]["source"], "ripe-rpki-validator-cache")
        self._test_option(self.cfg["rpki_roas"], "source", ("ripe-rpki-validator-cache","rtr"))
        self._test_mandatory(self.cfg["rpki_roas"], "source", has_default=True)

    def use_arin_whois_db_dump_enabled(self):
        """{}: use_arin_whois_db_dump.enabled"""
        self.assertEqual(self.cfg["filtering"]["irrdb"]["use_arin_whois_db_dump"]["enabled"], False)
        self._test_bool_val(self.cfg["filtering"]["irrdb"]["use_arin_whois_db_dump"], "enabled")
        self._test_mandatory(self.cfg["filtering"]["irrdb"]["use_arin_whois_db_dump"], "enabled", has_default=True)

    def use_registrobr_whois_db_dump_enabled(self):
        """{}: use_registrobr_whois_db_dump.enabled"""
        self.assertEqual(self.cfg["filtering"]["irrdb"]["use_registrobr_whois_db_dump"]["enabled"], False)
        self._test_bool_val(self.cfg["filtering"]["irrdb"]["use_registrobr_whois_db_dump"], "enabled")
        self._test_mandatory(self.cfg["filtering"]["irrdb"]["use_registrobr_whois_db_dump"], "enabled", has_default=True)

    def test_rpki_enabled(self):
        """{}: rpki_bgp_origin_validation, enabled"""
        self.assertEqual(self.cfg["filtering"]["rpki_bgp_origin_validation"]["enabled"], False)
        self._test_bool_val(self.cfg["filtering"]["rpki_bgp_origin_validation"], "enabled")
        self._test_mandatory(self.cfg["filtering"]["rpki_bgp_origin_validation"], "enabled", has_default=True)

    def test_rpki_reject_invalid(self):
        """{}: rpki_bgp_origin_validation, reject_invalid"""
        self.assertEqual(self.cfg["filtering"]["rpki_bgp_origin_validation"]["reject_invalid"], True)
        self._test_bool_val(self.cfg["filtering"]["rpki_bgp_origin_validation"], "reject_invalid")
        self._test_optional(self.cfg["filtering"]["rpki_bgp_origin_validation"], "reject_invalid")

    def test_reject_policy(self):
        """{}: reject_policy"""
        self.assertEqual(self.cfg["filtering"]["reject_policy"]["policy"], "reject")
        self._contains_err()

        self.cfg["communities"]["reject_cause"]["std"] = "65520:dyn_val"
        self.cfg["filtering"]["reject_policy"]["policy"] = "tag"
        self._contains_err()

        del self.cfg["communities"]["reject_cause"]["std"]
        self.cfg["filtering"]["reject_policy"]["policy"] = "tag"
        self._contains_err("The 'reject_cause' community must be configured when 'reject_policy.policy' is 'tag'.")

        self._test_option(self.cfg["filtering"]["reject_policy"], "policy", ())

        self._test_mandatory(self.cfg["filtering"]["reject_policy"], "policy", has_default=True)

    def test_blackhole_announce_to_client(self):
        """{}: blackhole_filtering, announce_to_client"""
        self.assertEqual(self.cfg["blackhole_filtering"]["announce_to_client"], True)
        self._test_bool_val(self.cfg["blackhole_filtering"], "announce_to_client")

    def test_blackhole_filtering_policy_ipv4(self):
        """{}: blackhole_filtering, policy_ipv4"""
        self.load_config(file_name="{}/test_cfg_general_blackhole_filtering.yml".format(os.path.dirname(__file__)))
        self.assertEqual(self.cfg["blackhole_filtering"]["policy_ipv4"], "propagate-unchanged")

        self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv4"] = "192.0.2.1"
        self._test_option(self.cfg["blackhole_filtering"], "policy_ipv4", ("propagate-unchanged", "rewrite-next-hop"))
        self._test_optional(self.cfg["blackhole_filtering"], "policy_ipv4")

        self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv4"] = ""
        self.cfg["blackhole_filtering"]["policy_ipv4"] = "propagate-unchanged"
        self._contains_err()
        self.cfg["blackhole_filtering"]["policy_ipv4"] = "rewrite-next-hop"
        self._contains_err("Since blackhole_filtering.policy_ipv4 is 'rewrite_next_hop', an IPv4 address must be provided in 'rewrite_next_hop_ipv4'.")

    def test_blackhole_filtering_policy_ipv6(self):
        """{}: blackhole_filtering, policy_ipv6"""
        self.load_config(file_name="{}/test_cfg_general_blackhole_filtering.yml".format(os.path.dirname(__file__)))
        self.assertEqual(self.cfg["blackhole_filtering"]["policy_ipv6"], "propagate-unchanged")

        self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv6"] = "fe80::1"
        self._test_option(self.cfg["blackhole_filtering"], "policy_ipv6", ("propagate-unchanged", "rewrite-next-hop"))
        self._test_optional(self.cfg["blackhole_filtering"], "policy_ipv6")

        self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv6"] = ""
        self.cfg["blackhole_filtering"]["policy_ipv6"] = "propagate-unchanged"
        self._contains_err()
        self.cfg["blackhole_filtering"]["policy_ipv6"] = "rewrite-next-hop"
        self._contains_err("Since blackhole_filtering.policy_ipv6 is 'rewrite_next_hop', an IPv6 address must be provided in 'rewrite_next_hop_ipv6'.")

    def test_blackhole_filtering_ipv4(self):
        """{}: blackhole_filtering, rewrite_next_hop, ipv4"""
        self.load_config(file_name="{}/test_cfg_general_blackhole_filtering.yml".format(os.path.dirname(__file__)))
        for ip in ("127.0.0.1", "192.168.0.1", "192.0.2.1", "12.34.56.78"):
            self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv4"] = ip
            self._contains_err()
        for ip in ("10.0.0.1/24", "a", "fe80::1"):
            self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv4"] = ip
            self._contains_err("Error parsing 'rewrite_next_hop_ipv4' at 'cfg.blackhole_filtering' level - Invalid IPv4 address: {}.".format(ip))
        self._test_optional(self.cfg["blackhole_filtering"], "rewrite_next_hop_ipv4")

    def test_graceful_shutdown(self):
        """{}: graceful shutdown"""
        self.assertEqual(self.cfg["graceful_shutdown"]["enabled"], False)
        self._test_bool_val(self.cfg["graceful_shutdown"], "enabled")

    def test_rfc1997_wellknown_communities(self):
        """{}: RFC1997 well-known communities"""
        self.assertEqual(self.cfg["rfc1997_wellknown_communities"]["policy"], "pass")
        self._test_option(self.cfg["rfc1997_wellknown_communities"], "policy", ("rfc1997", "pass"))

    def test_rtt_thresholds_str(self):
        """{}: RTT thresholds as comma separated string"""
        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  rtt_thresholds: '1, 2, 3'"
        )
        self._contains_err()

    def test_rtt_thresholds_int(self):
        """{}: RTT thresholds as list of int"""
        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  rtt_thresholds: 1, 2, 3"
        )
        self._contains_err()

    def test_rtt_thresholds_empty(self):
        """{}: RTT thresholds, empty"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  rtt_thresholds: "
        )
        self._contains_err()

    def test_rtt_thresholds_invalid(self):
        """{}: RTT thresholds, invalid values"""
        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  rtt_thresholds: 1, 2a, 3"
        )
        self._contains_err("RTT thresholds list items must be positive integers:  2a")

    def test_rtt_thresholds_order(self):
        """{}: RTT thresholds, out of order"""
        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  rtt_thresholds: 3, 2, 1"
        )
        self._contains_err("RTT thresholds list items must be provided in ascending order: 2 < 3")

    def test_blackhole_filtering_ipv6(self):
        """{}: blackhole_filtering, rewrite_next_hop, ipv6"""
        self.load_config(file_name="{}/test_cfg_general_blackhole_filtering.yml".format(os.path.dirname(__file__)))
        for ip in ("fe80::1", "2001:DB8::1"):
            self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv6"] = ip
            self._contains_err()
        for ip in ("10.0.0.1/24", "a"):
            self.cfg["blackhole_filtering"]["rewrite_next_hop_ipv6"] = ip
            self._contains_err("Error parsing 'rewrite_next_hop_ipv6' at 'cfg.blackhole_filtering' level - Invalid IPv6 address: {}.".format(ip))
        self._test_optional(self.cfg["blackhole_filtering"], "rewrite_next_hop_ipv6")

    def test_communities_std(self):
        """{}: standard BGP communities"""


        self.cfg["communities"]["blackholing"] = {}
        for c in self.VALID_STD_COMMS:
            self.cfg["communities"]["blackholing"]["std"] = c
            self._contains_err()
        for c in ("65536:666", "-1:-1", "0:-1", "65535:65536", "1", "1:1:1", "rt:1:0") + self.VALID_LRG_COMMS + self.VALID_EXT_COMMS:
            self.cfg["communities"]["blackholing"]["std"] = c
            self._contains_err("Invalid BGP standard community: {};".format(c))
        for c in ("65534:peer_as", "peer_as:1"):
            self.cfg["communities"]["blackholing"]["std"] = c
            self._contains_err("'peer_as' macro not allowed")

        c = "65535:1"
        self.cfg["communities"]["blackholing"]["std"] = c
        self._contains_err("range 65535:x is reserved")

        self.cfg["communities"]["blackholing"]["std"] = self.VALID_STD_COMMS[0]
        self._test_optional(self.cfg["communities"]["blackholing"], "std")

    def test_communities_lrg(self):
        """{}: large BGP communities"""

        for c in self.VALID_LRG_COMMS:
            self.cfg["communities"]["blackholing"]["lrg"] = c
            self._contains_err()
        for c in ("4294967296:1:1", "-1:0:0") + self.VALID_STD_COMMS + self.VALID_EXT_COMMS:
            self.cfg["communities"]["blackholing"]["lrg"] = c
            self._contains_err("Invalid BGP large community: {};".format(c))
        for c in ("1:65535:peer_as", "peer_as:1:1", "1:peer_as:2"):
            self.cfg["communities"]["blackholing"]["lrg"] = c
            self._contains_err("'peer_as' macro not allowed")
        self.cfg["communities"]["blackholing"]["lrg"] = self.VALID_LRG_COMMS[0]
        self._test_optional(self.cfg["communities"]["blackholing"], "lrg")

    def test_communities_ext(self):
        """{}: extended BGP communities"""

        for c in self.VALID_EXT_COMMS:
            self.cfg["communities"]["blackholing"]["ext"] = c
            self._contains_err()
        for c in ("x:1:0", "ro:-1:0") + self.VALID_STD_COMMS + self.VALID_LRG_COMMS:
            self.cfg["communities"]["blackholing"]["ext"] = c
            self._contains_err("Invalid BGP extended community: {};".format(c))
        for c in ("rt:65535:peer_as", "ro:peer_as:1"):
            self.cfg["communities"]["blackholing"]["ext"] = c
            self._contains_err("'peer_as' macro not allowed")
        self._test_optional(self.cfg["communities"]["blackholing"], "ext")

    def test_mandatory_peer_as_communities(self):
        """{}: communities that need peer_as macro"""

        for comm in ("announce_to_peer", "do_not_announce_to_peer",
                     "prepend_once_to_peer", "prepend_twice_to_peer",
                     "prepend_thrice_to_peer", "add_noexport_to_peer",
                     "add_noadvertise_to_peer"):
            for c in self.VALID_STD_COMMS:
                self.cfg["communities"][comm]["std"] = c
                self._contains_err("'peer_as' macro is mandatory in this community")
            self.cfg["communities"][comm]["std"] = None
            for c in self.VALID_LRG_COMMS:
                self.cfg["communities"][comm]["lrg"] = c
                self._contains_err("'peer_as' macro is mandatory in this community")
            self.cfg["communities"][comm]["lrg"] = None
            for c in self.VALID_EXT_COMMS:
                self.cfg["communities"][comm]["ext"] = c
                self._contains_err("'peer_as' macro is mandatory in this community")
            self.cfg["communities"][comm]["ext"] = None

    def test_mandatory_dyn_val_communities(self):
        """{}: communities that need dyn_val macro"""

        for comm in ("reject_cause", "rejected_route_announced_by",
                     "do_not_announce_to_peers_with_rtt_lower_than",
                     "do_not_announce_to_peers_with_rtt_higher_than",
                     "announce_to_peers_with_rtt_lower_than",
                     "announce_to_peers_with_rtt_higher_than",
                     "prepend_once_to_peers_with_rtt_lower_than",
                     "prepend_twice_to_peers_with_rtt_lower_than",
                     "prepend_thrice_to_peers_with_rtt_lower_than",
                     "prepend_once_to_peers_with_rtt_higher_than",
                     "prepend_twice_to_peers_with_rtt_higher_than",
                     "prepend_thrice_to_peers_with_rtt_higher_than"):
            for c in self.VALID_STD_COMMS:
                self.cfg["communities"][comm]["std"] = c
                self._contains_err("'dyn_val' macro is mandatory in this community")
            self.cfg["communities"][comm]["std"] = None
            for c in self.VALID_LRG_COMMS:
                self.cfg["communities"][comm]["lrg"] = c
                self._contains_err("'dyn_val' macro is mandatory in this community")
            self.cfg["communities"][comm]["lrg"] = None
            for c in self.VALID_EXT_COMMS:
                self.cfg["communities"][comm]["ext"] = c
                self._contains_err("'dyn_val' macro is mandatory in this community")
            self.cfg["communities"][comm]["ext"] = None

    def test_reject_cause_community_with_no_reject_policy(self):
        """{}: reject_cause can be set only with 'tag' reject_policy"""
        self.cfg["communities"]["reject_cause"]["std"] = "0:dyn_val"
        self._contains_err("The 'reject_cause' community can be set only if 'reject_policy.policy' is 'tag'.")

        self.cfg["filtering"]["reject_policy"]["policy"] = "tag"
        self._contains_err()

    def test_rejected_route_announced_by_with_no_reject_policy(self):
        """{}: rejected_route_announced_by can be set only with 'tag' reject_policy"""
        self.cfg["communities"]["reject_cause"]["std"] = "65520:dyn_val"
        self.cfg["communities"]["rejected_route_announced_by"]["std"] = "0:dyn_val"
        self._contains_err("The 'rejected_route_announced_by' community can be set only if 'reject_policy.policy' is 'tag'.")

        self.cfg["filtering"]["reject_policy"]["policy"] = "tag"
        self._contains_err()

    def test_peer_as_usage_in_communities(self):
        """{}: peer_as macro usage in communities"""

        comm = "announce_to_peer"

        self.cfg["communities"][comm]["std"] = "rs_as:peer_as"
        self._contains_err()
        self.cfg["communities"][comm]["std"] = None

        self.cfg["communities"][comm]["std"] = "peer_as:rs_as"
        self._contains_err("'peer_as' macro can be used only in the last part of the value")
        self.cfg["communities"][comm]["std"] = None

        self.cfg["communities"][comm]["lrg"] = "rs_as:rs_as:peer_as"
        self._contains_err()
        self.cfg["communities"][comm]["lrg"] = None

        self.cfg["communities"][comm]["lrg"] = "peer_as:rs_as:0"
        self._contains_err("'peer_as' macro can be used only in the last part of the value")
        self.cfg["communities"][comm]["lrg"] = None

        self.cfg["communities"][comm]["lrg"] = "rs_as:peer_as:0"
        self._contains_err("'peer_as' macro can be used only in the last part of the value")
        self.cfg["communities"][comm]["lrg"] = None

    def test_dyn_val_usage_in_communities(self):
        """{}: dyn_val macro usage in communities"""

        comm = "reject_cause"

        self.cfg["filtering"]["reject_policy"]["policy"] = "tag"

        self.cfg["communities"][comm]["std"] = "rs_as:dyn_val"
        self._contains_err()
        self.cfg["communities"][comm]["std"] = None

        self.cfg["communities"][comm]["std"] = "dyn_val:rs_as"
        self._contains_err("'dyn_val' macro can be used only in the last part of the value")
        self.cfg["communities"][comm]["std"] = None

        self.cfg["communities"][comm]["lrg"] = "rs_as:rs_as:dyn_val"
        self._contains_err()
        self.cfg["communities"][comm]["lrg"] = None

        self.cfg["communities"][comm]["lrg"] = "dyn_val:rs_as:0"
        self._contains_err("'dyn_val' macro can be used only in the last part of the value")
        self.cfg["communities"][comm]["lrg"] = None

        self.cfg["communities"][comm]["lrg"] = "rs_as:dyn_val:0"
        self._contains_err("'dyn_val' macro can be used only in the last part of the value")
        self.cfg["communities"][comm]["lrg"] = None

    def test_custom_communities_valid(self):
        """{}: custom communities: valid"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
        ]
        yaml_lines = tpl + [
            "  custom_communities:",
            "    test1:",
            "      std: '1:1'",
            "    test2:",
            "      lrg: '2:2:2'",
            "    test3:",
            "      std: '3:3'",
            "      lrg: '3:3:3'",
            "      ext: 'rt:3:3'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err()

    def test_custom_communities_reserved(self):
        """{}: custom communities: reserved name"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
        ]
        yaml_lines = tpl + [
            "  custom_communities:",
            "    blackholing:",
            "      std: '1:1'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("The custom community name 'blackholing' collides with a built-in community with the same name.")

    def test_custom_communities_invalid(self):
        """{}: custom communities: invalid"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
        ]
        yaml_lines = tpl + [
            "  custom_communities:",
            "    test1:",
            "      std: peer_as"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Error parsing 'std' at 'cfg.custom_communities.test1' level - Invalid BGP standard community: peer_as; it must be in the x:x format")

        yaml_lines = tpl + [
            "  custom_communities:",
            "    test1:",
            "      std: '1:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Error parsing 'std' at 'cfg.custom_communities.test1' level - Invalid BGP standard community: 1:peer_as")

    def test_duplicate_communities(self):
        """{}: duplicate communities"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:",
            "    prefix_present_in_as_set:",
            "      std: '999:1'",
            "    prefix_not_present_in_as_set:",
            "      lrg: 'rs_as:2:2'",
        ]
        yaml_lines = tpl + [
            "    origin_present_in_as_set:",
            "      std: 'rs_as:1'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("The 'prefix_present_in_as_set.std' community's value (999:1) has already been used for another community.")

        yaml_lines = tpl + [
            "    origin_present_in_as_set:",
            "      lrg: '999:2:2'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("The 'prefix_not_present_in_as_set.lrg' community's value (999:2:2) has already been used for another community.")

        yaml_lines = tpl + [
            "  custom_communities:",
            "    test:",
            "      std: '999:1'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("The 'test.std' community's value (999:1) has already been used for another community.")

    def test_overlapping_communities_internal(self):
        """{}: overlapping communities, internal"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:",
            "    reject_cause:",
            "      std: 1:dyn_val"
        ]

        # internal/outbound (prefix_not_present_in_as_set)
        yaml_lines = tpl + [
            "    prefix_not_present_in_as_set:",
            "      std: '1:1'",
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'reject_cause' and 'prefix_not_present_in_as_set' overlap: 1:dyn_val / 1:1. Internal communities can't have overlapping values with any other community.")

        # internal/inbound (do_not_announce_to_peer)
        yaml_lines = tpl + [
            "    do_not_announce_to_peer:",
            "      std: '1:peer_as'",
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'reject_cause' and 'do_not_announce_to_peer' overlap: 1:dyn_val / 1:peer_as. Internal communities can't have overlapping values with any other community.")

        # internal/custom
        yaml_lines = tpl + [
            "  custom_communities:",
            "    test:",
            "      std: '1:0'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'reject_cause' and 'test' overlap: 1:dyn_val / 1:0. Internal communities can't have overlapping values with any other community.")

    def test_overlapping_communities_out_in(self):
        """{}: overlapping communities, outbound/inbound"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:"
        ]

        # prefix_not_present_in_as_set: outbound
        # do_not_announce_to_peer: inbound
        yaml_lines = tpl + [
            "    prefix_not_present_in_as_set:",
            "      std: '0:1'",
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'do_not_announce_to_peer' and 'prefix_not_present_in_as_set' overlap: 0:peer_as / 0:1. Inbound communities and outbound communities can't have overlapping values, otherwise they might be scrubbed.")

        # Same as above, but with 0 in the prefix_not_present_in_as_set comm.
        # No errors because 'peer_as' can't be 0.
        yaml_lines = tpl + [
            "    prefix_not_present_in_as_set:",
            "      std: '0:0'",
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err()

        # Same as above, but with 'rs_as' in the prefix_not_present_in_as_set comm.
        # No errors because 'peer_as' can't be the route server's ASN itself.
        yaml_lines = tpl + [
            "    prefix_not_present_in_as_set:",
            "      std: '0:rs_as'",
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err()

        # Same as above, but with a private ASN in the prefix_not_present_in_as_set comm.
        # No errors because 'peer_as' can't be a private ASN.
        yaml_lines = tpl + [
            "    prefix_not_present_in_as_set:",
            "      std: '0:65501'",
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err()

        # Testing with allow_private_asns=False...
        with self.assertRaises(ConfigError):
            self.cfg.check_overlapping_communities(allow_private_asns=False)
        exp_err_msg_found = False
        for line in self.logger_handler.msgs:
            if "Community 'do_not_announce_to_peer' and 'prefix_not_present_in_as_set' overlap: 0:peer_as / 0:65501. Inbound communities and outbound communities can't have overlapping values, otherwise they might be scrubbed." in line:
                exp_err_msg_found = True
                break

        if not exp_err_msg_found:
            self.fail("Expected error message not found")

    def test_overlapping_communities_in_in(self):
        """{}: overlapping communities, inbound/inbound"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:"
        ]

        # blackholing: inbound
        # do_not_announce_to_peer: inbound
        yaml_lines = tpl + [
            "    blackholing:",
            "      std: '0:666'",
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'blackholing' and 'do_not_announce_to_peer' overlap: 0:666 / 0:peer_as. Inbound communities can't have overlapping values, otherwise their meaning could be uncertain.")

        # Same as above, but with 0 in the last part of blackholing
        # No errors because 'peer_as' can't be 0.
        yaml_lines = tpl + [
            "    blackholing:",
            "      std: '0:0'",
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err()

        # Same as above, but with rs_as in the last part of blackholing
        # No errors because 'peer_as' can't be the route server's ASN itself.
        yaml_lines = tpl + [
            "    blackholing:",
            "      std: '0:rs_as'",
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err()

        # Same as above, but with a private ASN in the last part of
        # blackholing community.
        yaml_lines = tpl + [
            "    blackholing:",
            "      std: '0:65501'",
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err()

        # Testing with allow_private_asns=False...
        # It must work, peer_as can't be a private ASN; moreover
        # also on OpenBGPD (where inbound communities are scrubbed
        # using a wildcard) when the 'x:peer_as' community is
        # deleted also the 'x:<asn>' is.
        self.cfg.check_overlapping_communities(allow_private_asns=False)
        self._contains_err()

    def test_overlapping_communities_in_in_dyn_val(self):
        """{}: overlapping communities, inbound/inbound (dyn_val)"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:"
        ]

        # blackholing: inbound
        # do_not_announce_to_peers_with_rtt_lower_than: inbound with dyn_val
        yaml_lines = tpl + [
            "    blackholing:",
            "      std: '0:666'",
            "    do_not_announce_to_peers_with_rtt_lower_than:",
            "      std: '0:dyn_val'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'blackholing' and 'do_not_announce_to_peers_with_rtt_lower_than' overlap: 0:666 / 0:dyn_val. Inbound communities can't have overlapping values, otherwise their meaning could be uncertain.")

        # do_not_announce_to_peer: inbound with peer_as
        # do_not_announce_to_peers_with_rtt_lower_than: inbound with dyn_val
        yaml_lines = tpl + [
            "    do_not_announce_to_peer:",
            "      std: '0:peer_as'",
            "    do_not_announce_to_peers_with_rtt_lower_than:",
            "      std: '0:dyn_val'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'do_not_announce_to_peer' and 'do_not_announce_to_peers_with_rtt_lower_than' overlap: 0:peer_as / 0:dyn_val. Inbound communities can't have overlapping values, otherwise their meaning could be uncertain.")

    def test_overlapping_communities_out_in_dyn_val(self):
        """{}: overlapping communities, outbound/inbound (dyn_val)"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:"
        ]

        # prefix_not_present_in_as_set: outbound
        # do_not_announce_to_peers_with_rtt_lower_than: inbound with dyn_val
        yaml_lines = tpl + [
            "    prefix_not_present_in_as_set:",
            "      std: '0:1'",
            "    do_not_announce_to_peers_with_rtt_lower_than:",
            "      std: '0:dyn_val'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'do_not_announce_to_peers_with_rtt_lower_than' and 'prefix_not_present_in_as_set' overlap: 0:dyn_val / 0:1. Inbound communities and outbound communities can't have overlapping values, otherwise they might be scrubbed.")

    def test_overlapping_communities_in_cust(self):
        """{}: overlapping communities, inbound/custom"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:"
        ]

        # Custom community overlaps
        yaml_lines = tpl + [
            "    do_not_announce_to_peer:",
            "      std: '1:peer_as'",
            "  custom_communities:",
            "    test1:",
            "      std: '1:1'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err("Community 'do_not_announce_to_peer' and 'test1' overlap: 1:peer_as / 1:1. Inbound communities and custom communities can't have overlapping values, otherwise they might be scrubbed.")

        # Private ASNs
        yaml_lines = tpl + [
            "    do_not_announce_to_peer:",
            "      std: '1:peer_as'",
            "  custom_communities:",
            "    test1:",
            "      std: '1:65501'"
        ]
        self.load_config(yaml="\n".join(yaml_lines))
        self._contains_err()

        # Testing with allow_private_asns=False...
        with self.assertRaises(ConfigError):
            self.cfg.check_overlapping_communities(allow_private_asns=False)
        exp_err_msg_found = False
        for line in self.logger_handler.msgs:
            if "Community 'do_not_announce_to_peer' and 'test1' overlap: 1:peer_as / 1:65501. Inbound communities and custom communities can't have overlapping values, otherwise they might be scrubbed." in line:
                exp_err_msg_found = True
                break

        if not exp_err_msg_found:
            self.fail("Expected error message not found")

    def test_rtt_based_communities_without_rtt_thresholds(self):
        """{}: RTT-based communities without RTT thresholds"""
        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  communities:",
            "    do_not_announce_to_peers_with_rtt_lower_than:",
            "      std: '0:dyn_val'"
        ]
        self.load_config(yaml="\n".join(tpl))
        self._contains_err("Some RTT-based functions are configured but the RTT thresholds list is empty.")

        tpl = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  rtt_thresholds: 1, 2",
            "  communities:",
            "    do_not_announce_to_peers_with_rtt_lower_than:",
            "      std: '0:dyn_val'"
        ]
        self.load_config(yaml="\n".join(tpl))
        self._contains_err()

    def test_max_pref_action(self):
        """{}: max_prefix action"""
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["action"], None)
        self._test_option(self.cfg["filtering"]["max_prefix"], "action", ("shutdown", "restart", "block", "warning"))
        self._test_optional(self.cfg["filtering"]["max_prefix"], "action")

    def test_max_pref_peeringdb(self):
        """{}: max_prefix PeeringDB"""
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["peering_db"], {
            "enabled": True,
            "increment": {
                "absolute": 100,
                "relative": 15
            }
        })
        self._test_bool_val(self.cfg["filtering"]["max_prefix"]["peering_db"], "enabled")
        self._test_optional(self.cfg["filtering"]["max_prefix"]["peering_db"], "enabled")

    def test_max_pref_peeringdb_pre_0_13_0(self):
        """{}: max_prefix PeeringDB (pre v0.13.0 format)"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  filtering:\n"
            "    max_prefix:\n"
            "      peering_db: True\n"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["filtering"]["max_prefix"]["peering_db"]["enabled"], True)
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["peering_db"]["increment"]["absolute"], 100)
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["peering_db"]["increment"]["relative"], 15)

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2\n"
            "  filtering:\n"
            "    max_prefix:\n"
            "      peering_db: False\n"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["filtering"]["max_prefix"]["peering_db"]["enabled"], False)

    def test_max_pref_general_limit_ipv4(self):
        """{}: max_prefix general_limit_ipv4"""
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["general_limit_ipv4"], 170000)
        self._test_optional(self.cfg["filtering"]["max_prefix"], "general_limit_ipv4")

    def test_max_pref_general_limit_ipv6(self):
        """{}: max_prefix general_limit_ipv6"""
        self.assertEqual(self.cfg["filtering"]["max_prefix"]["general_limit_ipv6"], 12000)
        self._test_optional(self.cfg["filtering"]["max_prefix"], "general_limit_ipv6")

    def test_transit_free_action(self):
        """{}: transit free, action"""
        self.assertEqual(self.cfg["filtering"]["transit_free"]["action"], None)
        self._test_option(self.cfg["filtering"]["transit_free"], "action", ("reject", "warning"))
        self._test_optional(self.cfg["filtering"]["transit_free"], "action")

    def test_never_via_route_servers_peering_db(self):
        """{}: never via route-servers, peering_db"""
        self.assertEqual(self.cfg["filtering"]["never_via_route_servers"]["peering_db"], True)
        self._test_bool_val(self.cfg["filtering"]["never_via_route_servers"], "peering_db")

    def test_never_via_route_servers_asns(self):
        """{}: never via route-servers, asns"""
        self.assertEqual(self.cfg["filtering"]["never_via_route_servers"]["asns"], None)
        self._test_optional(self.cfg["filtering"]["never_via_route_servers"], "asns")

    def test_transit_free_asns(self):
        """{}: transit free, ASNs list"""
        self.assertEqual(self.cfg["filtering"]["transit_free"]["asns"], [174, 209, 286, 701, 1239, 1299, 2828, 2914, 3257, 3320, 3356, 3549, 5511, 6453, 6461, 6762, 6830, 7018, 12956])
        self._test_optional(self.cfg["filtering"]["transit_free"], "asns")

        cfg = [
            "cfg:",
            "  rs_as: 999",
            "  router_id: 192.0.2.2",
            "  filtering:",
            "    transit_free:",
        ]
        self.load_config(yaml="\n".join(cfg + [
            "      action: reject"
        ]))
        self._contains_err()
        self.assertEqual(self.cfg["filtering"]["transit_free"]["action"], "reject")

        self.load_config(yaml="\n".join(cfg + [
            "      action: test"
        ]))
        self._contains_err("Error parsing 'action' at 'cfg.filtering.transit_free' level - Invalid option for 'action':")

        self.load_config(yaml="\n".join(cfg + [
            "      action: reject",
            "      asns: '1, 2, 3, 4'"
        ]))
        self._contains_err()
        self.assertEqual(self.cfg["filtering"]["transit_free"]["action"], "reject")
        self.assertEqual(self.cfg["filtering"]["transit_free"]["asns"], [1,2,3,4])

        self.load_config(yaml="\n".join(cfg + [
            "      action: reject",
            "      asns: '1, 2, 3, 4a'"
        ]))
        self._contains_err("Error parsing 'asns' at 'cfg.filtering.transit_free' level - Invalid ASN:  4a.")

    def test_default_values(self):
        """{}: minimal config"""
        self.load_config(yaml="cfg:\n"
            "  rs_as: 999\n"
            "  router_id: 192.0.2.2"
        )
        self._contains_err()

        exp_res = {
            "rs_as": 999,
            "router_id": "192.0.2.2",
            "prepend_rs_as": False,
            "path_hiding": True,
            "passive": True,
            "gtsm": False,
            "add_path": False,
            "filtering": {
                "next_hop": {
                    "policy": "strict"
                },
                "global_black_list_pref": None,
                "ipv4_pref_len": {
                    "min": 8,
                    "max": 24
                },
                "ipv6_pref_len": {
                    "min": 12,
                    "max": 48
                },
                "max_as_path_len": 32,
                "reject_invalid_as_in_as_path": True,
                "reject_policy": {
                    "policy": "reject"
                },
                "transit_free": {
                    "action": None,
                    "asns": None
                },
                "never_via_route_servers": {
                    "peering_db": True,
                    "asns": None
                },
                "irrdb": {
                    "tag_as_set": True,
                    "enforce_origin_in_as_set": True,
                    "enforce_prefix_in_as_set": True,
                    "allow_longer_prefixes": False,
                    "peering_db": False,
                    "use_rpki_roas_as_route_objects": {
                        "enabled": False,
                    },
                    "use_arin_bulk_whois_data": {
                        "enabled": False,
                        "source": "http://irrexplorer.nlnog.net/static/dumps/arin-whois-originas.json.bz2"
                    },
                    "use_registrobr_bulk_whois_data": {
                        "enabled": False,
                        "source": "ftp://ftp.registro.br/pub/numeracao/origin/nicbr-asn-blk-latest.txt"
                    },
                },
                "rpki_bgp_origin_validation": {
                    "enabled": False,
                    "reject_invalid": True,
                },
                "max_prefix": {
                    "action": None,
                    "restart_after": 15,
                    "peering_db": {
                        "enabled": True,
                        "increment": {
                            "absolute": 100,
                            "relative": 15
                        }
                    },
                    "general_limit_ipv4": 170000,
                    "general_limit_ipv6": 12000
                },
            },
            "rtt_thresholds": None,
            "rpki_roas": {
                "source": "ripe-rpki-validator-cache",
                "ripe_rpki_validator_url": [
                    "https://rpki-validator.ripe.net/api/export.json",
                    "https://rpki.gin.ntt.net/api/export.json"
                ],
                "allowed_trust_anchors": [
                    "APNIC RPKI Root",
                    "AfriNIC RPKI Root",
                    "LACNIC RPKI Root",
                    "RIPE NCC RPKI Root",
                    "apnic",
                    "afrinic",
                    "lacnic",
                    "ripe"
                ]
            },
            "blackhole_filtering": {
                "policy_ipv4": None,
                "policy_ipv6": None,
                "rewrite_next_hop_ipv4": None,
                "rewrite_next_hop_ipv6": None,
                "announce_to_client": True,
                "add_noexport": True,
            },
            "graceful_shutdown": {
                "enabled": False,
                "local_pref": 0
            },
            "rfc1997_wellknown_communities": {
                "policy": "pass"
            }
        }

        self.maxDiff = None
        del self.cfg["communities"]
        del self.cfg["custom_communities"]
        self.assertMultiLineEqual(
            yaml.safe_dump(self.cfg.cfg, default_flow_style=False),
            yaml.safe_dump({"cfg": exp_res}, default_flow_style=False)
        )

    def test_distrib_config(self):
        """{}: distributed config"""
        self.load_config(file_name="config.d/general.yml")
        self.cfg.parse()
        self._contains_err()

        exp_res = {
            "rs_as": 999,
            "router_id": "192.0.2.2",
            "prepend_rs_as": False,
            "path_hiding": True,
            "passive": True,
            "gtsm": False,
            "add_path": False,
            "filtering": {
                "next_hop": {
                    "policy": "strict"
                },
                "global_black_list_pref": None,
                "ipv4_pref_len": {
                    "min": 8,
                    "max": 24
                },
                "ipv6_pref_len": {
                    "min": 12,
                    "max": 48
                },
                "max_as_path_len": 32,
                "reject_invalid_as_in_as_path": True,
                "reject_policy": {
                    "policy": "reject"
                },
                "transit_free": {
                    "action": None,
                    "asns": [174, 209, 286, 701, 1239, 1299, 2828, 2914, 3257, 3320, 3356, 3549, 5511, 6453, 6461, 6762, 6830, 7018, 12956]
                },
                "never_via_route_servers": {
                    "peering_db": True,
                    "asns": None
                },
                "irrdb": {
                    "tag_as_set": True,
                    "enforce_origin_in_as_set": True,
                    "enforce_prefix_in_as_set": True,
                    "allow_longer_prefixes": False,
                    "peering_db": False,
                    "use_rpki_roas_as_route_objects": {
                        "enabled": False,
                    },
                    "use_arin_bulk_whois_data": {
                        "enabled": False,
                        "source": "http://irrexplorer.nlnog.net/static/dumps/arin-whois-originas.json.bz2"
                    },
                    "use_registrobr_bulk_whois_data": {
                        "enabled": False,
                        "source": "ftp://ftp.registro.br/pub/numeracao/origin/nicbr-asn-blk-latest.txt"
                    }
                },
                "rpki_bgp_origin_validation": {
                    "enabled": False,
                    "reject_invalid": True,
                },
                "max_prefix": {
                    "action": None,
                    "restart_after": 15,
                    "peering_db": {
                        "enabled": True,
                        "increment": {
                            "absolute": 100,
                            "relative": 15
                        }
                    },
                    "general_limit_ipv4": 170000,
                    "general_limit_ipv6": 12000
                },
            },
            "rtt_thresholds": [5, 10, 15, 20, 30, 50, 100, 200, 500],
            "rpki_roas": {
                "source": "ripe-rpki-validator-cache",
                "ripe_rpki_validator_url": [
                    "https://rpki-validator.ripe.net/api/export.json",
                    "https://rpki.gin.ntt.net/api/export.json"
                ],
                "allowed_trust_anchors": [
                    "APNIC RPKI Root",
                    "AfriNIC RPKI Root",
                    "LACNIC RPKI Root",
                    "RIPE NCC RPKI Root",
                    "apnic",
                    "afrinic",
                    "lacnic",
                    "ripe"
                ]
            },
            "blackhole_filtering": {
                "policy_ipv4": None,
                "policy_ipv6": None,
                "rewrite_next_hop_ipv4": None,
                "rewrite_next_hop_ipv6": None,
                "announce_to_client": True,
                "add_noexport": True,
            },
            "graceful_shutdown": {
                "enabled": False,
                "local_pref": 0
            },
            "rfc1997_wellknown_communities": {
                "policy": "pass"
            }
        }

        self.maxDiff = None
        del self.cfg["communities"]
        del self.cfg["custom_communities"]
        self.assertMultiLineEqual(
            yaml.safe_dump(self.cfg.cfg, default_flow_style=False),
            yaml.safe_dump({"cfg": exp_res}, default_flow_style=False)
        )

    def test_deprecated_rpki_validation(self):
        """{}: deprecated syntax, RPKI Origin Validation"""
        cfg = {
            "cfg": {
                "rs_as": 999,
                "router_id": "192.0.2.2",
                "filtering": {
                    "rpki": {
                        "enabled": True
                    }
                }
            }
        }
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertEqual(self.cfg.cfg["cfg"]["filtering"]["rpki_bgp_origin_validation"]["enabled"], True)

        cfg["cfg"]["filtering"]["rpki_bgp_origin_validation"] = {}
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err("A conflict due to a deprecated syntax exists: filtering.rpki and filtering.rpki_bgp_origin_validation are both configured.")

    def test_deprecated_ripe_rpki_validator_url(self):
        """{}: deprecated syntax, RPKI ROAs cache - multiple URLs"""
        cfg = {
            "cfg": {
                "rs_as": 999,
                "router_id": "192.0.2.2",
                "rpki_roas": {
                    "ripe_rpki_validator_url": "Foo"
                }
            }
        }
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["source"], "ripe-rpki-validator-cache")
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["ripe_rpki_validator_url"], ["Foo"])

    def test_deprecated_rpki_roas_source_rtrlib(self):
        """{}: deprecated syntax, RPKI ROAs source: rtrlib"""
        cfg = {
            "cfg": {
                "rs_as": 999,
                "router_id": "192.0.2.2",
                "filtering": {
                    "irrdb": {
                        "use_rpki_roas_as_route_objects": {
                            "enabled": True
                        }
                    },
                },
                "rpki_roas": {
                    "source": "rtrlib"
                }
            }
        }
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["source"], "rtr")

    def test_deprecated_rpki_roas_source(self):
        """{}: deprecated syntax, RPKI ROAs source"""
        cfg = {
            "cfg": {
                "rs_as": 999,
                "router_id": "192.0.2.2",
                "filtering": {
                    "irrdb": {
                        "use_rpki_roas_as_route_objects": {
                            "enabled": True
                        }
                    },
                    "rpki": {
                        "enabled": True
                    }
                }
            }
        }
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["source"], "rtr")

        cfg["cfg"]["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["source"] = "ripe-rpki-validator-cache"
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err("A deprecated syntax triggered an issue with the configuration of RPKI BGP Origin Validation (filtering.rpki) and ROAs-as-route-objects (filtering.irrdb.rpki_roas_as_route_objects). The former uses RTR as source for ROAs, while the latter is configured to use the RIPE RPKI Validator")

        del cfg["cfg"]["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["source"]
        cfg["cfg"]["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["enabled"] = False
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["source"], "rtr")

        cfg["cfg"]["filtering"]["rpki"]["enabled"] = False
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertTrue("rpki_roas" not in self.cfg.cfg)

        cfg["cfg"]["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["enabled"] = True
        cfg["cfg"]["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["source"] = "ripe-rpki-validator-cache"
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["source"], "ripe-rpki-validator-cache")

        cfg["cfg"]["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["ripe_rpki_validator_url"] = "Foo"
        cfg["cfg"]["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["allowed_trust_anchors"] = ["bar"]
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["source"], "ripe-rpki-validator-cache")
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["ripe_rpki_validator_url"], ["Foo"])
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["allowed_trust_anchors"], ["bar"])

        cfg["cfg"]["filtering"]["irrdb"]["use_rpki_roas_as_route_objects"]["ripe_rpki_validator_url"] = ["Foo", "Bar"]
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err()
        self.assertEqual(self.cfg.cfg["cfg"]["rpki_roas"]["ripe_rpki_validator_url"], ["Foo", "Bar"])

        cfg["cfg"]["rpki_roas"] = {}
        self.load_config(yaml=yaml.dump(cfg))
        self._contains_err("A conflict due to a deprecated syntax exists: please check rpki_roas, filtering.rpki and filtering.irrdb.rpki_roas_as_route_objects.")

    @mock.patch.dict(os.environ, {"ROUTER_ID": "192.0.2.1"})
    def test_env_vars_ok(self):
        """{}: environment variables: ok"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: ${ROUTER_ID}\n"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["router_id"], "192.0.2.1")

    @mock.patch.dict(os.environ, {"ROUTER_ID": "192.0.2.1"})
    def test_env_vars_missing(self):
        """{}: environment variables: ok"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: ${ROUTER_ID}\n"
            "  filtering:\n"
            "    global_black_list_pref: ${GLOBAL_BLACK_LIST_PREF}\n"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["router_id"], "192.0.2.1")
        self.assertEqual(self.cfg["filtering"]["global_black_list_pref"], None)

    @mock.patch.dict(os.environ, {"ROUTER_ID": "192.0.2.1"})
    def test_env_vars_corrupted(self):
        """{}: environment variables: corrupted"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: ${ROUTER_ID\n"
        )
        self._contains_err("Invalid IPv4 address: ${ROUTER_ID")

    @mock.patch.dict(os.environ, {"ROUTER_ID": "192.0.2.1", "BAD_ESCAPE": r"\u"})
    def test_env_vars_bad_escape(self):
        """{}: environment variables: ok (bad escape)"""

        self.cfg._load_from_yaml(
            "cfg:\n"
            "  rs_as: 999\n"
            "  router_id: ${ROUTER_ID}\n"
            "  filtering:\n"
            "    global_black_list_pref: ${GLOBAL_BLACK_LIST_PREF}\n"
        )
        self.cfg.parse()
        self._contains_err()

        self.assertEqual(self.cfg["router_id"], "192.0.2.1")
        self.assertEqual(self.cfg["filtering"]["global_black_list_pref"], None)



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import six
import unittest

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPDInstance, \
                                                          OpenBGPDPreviousInstance, \
                                                          OpenBGPDLatestInstance
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance

class BGPCommunitiesScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance(),

            cls.CLIENT_INSTANCE_CLASS(
                "AS1",
                cls.DATA["AS1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS131073",
                cls.DATA["AS131073_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS131073.j2"),
                        "/etc/bird/bird.conf"
                    )
                ]
            )
        ]

    def set_instance_variables(self):
        self.AS1 = self._get_instance_by_name("AS1")
        self.AS2 = self._get_instance_by_name("AS2")
        self.AS131073 = self._get_instance_by_name("AS131073")
        self.rs = self._get_instance_by_name("rs")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1)
        self.session_is_up(self.rs, self.AS2)
        self.session_is_up(self.rs, self.AS131073)

    def test_031_only_to_AS1_std(self):
        """{}: announce to AS1 only (std)"""
        pref = self.DATA["AS2_only_to_AS1_s"]
        self.receive_route(self.rs, pref, self.AS2,
                           std_comms=["0:999", "999:1"],
                           ext_comms=[],
                           lrg_comms=[])
        self.receive_route(self.AS1, pref, self.rs,
                           std_comms=[], ext_comms=[], lrg_comms=[])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS131073, pref)
        msg = ("route didn't pass control communities checks - "
               "NOT ANNOUNCING {} TO {{inst}}".format(pref))
        self.log_contains(self.rs, msg, {"inst": self.AS131073})

    def test_031_only_to_AS1_ext(self):
        """{}: announce to AS1 only (ext)"""
        pref = self.DATA["AS2_only_to_AS1_e"]
        self.receive_route(self.rs, pref, self.AS2,
                           std_comms=[],
                           ext_comms=["rt:0:999", "rt:999:1"],
                           lrg_comms=[])
        if isinstance(self.rs, OpenBGPDInstance):
            #TODO: don't check for ext_comms on OpenBGPD since
            #      it seems not possible to remove ext comms using
            #      wildcard, so ext comms scrubbing is not implemented
            self.receive_route(self.AS1, pref, self.rs,
                               std_comms=[], lrg_comms=[])
        else:
            self.receive_route(self.AS1, pref, self.rs,
                               std_comms=[], ext_comms=[], lrg_comms=[])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS131073, pref)
        msg = ("route didn't pass control communities checks - "
               "NOT ANNOUNCING {} TO {{inst}}".format(pref))
        self.log_contains(self.rs, msg, {"inst": self.AS131073})

    def test_031_only_to_AS1_lrg(self):
        """{}: announce to AS1 only (lrg)"""
        pref = self.DATA["AS2_only_to_AS1_l"]
        self.receive_route(self.rs, pref, self.AS2,
                           std_comms=[],
                           ext_comms=[],
                           lrg_comms=["999:0:999", "999:999:1"])
        self.receive_route(self.AS1, pref, self.rs,
                           std_comms=[], ext_comms=[], lrg_comms=[])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS131073, pref)
        msg = ("route didn't pass control communities checks - "
               "NOT ANNOUNCING {} TO {{inst}}".format(pref))
        self.log_contains(self.rs, msg, {"inst": self.AS131073})

    def test_032_only_to_AS131073_ext(self):
        """{}: announce to AS131073 only (ext)"""
        pref = self.DATA["AS2_only_to_AS131073_e"]
        self.receive_route(self.rs, pref, self.AS2,
                           std_comms=["0:999"],
                           ext_comms=["rt:999:131073"],
                           lrg_comms=[])
        if isinstance(self.rs, OpenBGPDInstance):
            #TODO: don't check for ext_comms on OpenBGPD since
            #      it seems not possible to remove ext comms using
            #      wildcard, so ext comms scrubbing is not implemented
            self.receive_route(self.AS131073, pref, self.rs,
                            std_comms=[], lrg_comms=[])
        else:
            self.receive_route(self.AS131073, pref, self.rs,
                            std_comms=[], ext_comms=[], lrg_comms=[])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS1, pref)
        msg = ("route didn't pass control communities checks - "
               "NOT ANNOUNCING {} TO {{inst}}".format(pref))
        self.log_contains(self.rs, msg, {"inst": self.AS1})

    def test_032_only_to_AS131073_lrg(self):
        """{}: announce to AS131073 only (lrg)"""
        pref = self.DATA["AS2_only_to_AS131073_l"]
        self.receive_route(self.rs, pref, self.AS2,
                           std_comms=[],
                           ext_comms=[],
                           lrg_comms=["999:0:999", "999:999:131073"])
        if isinstance(self.rs, OpenBGPDInstance):
            #TODO: don't check for ext_comms on OpenBGPD since
            #      it seems not possible to remove ext comms using
            #      wildcard, so ext comms scrubbing is not implemented
            self.receive_route(self.AS131073, pref, self.rs,
                               std_comms=[], lrg_comms=[])
        else:
            self.receive_route(self.AS131073, pref, self.rs,
                               std_comms=[], ext_comms=[], lrg_comms=[])
        with six.assertRaisesRegex(self, AssertionError, "Routes not found."):
            self.receive_route(self.AS1, pref)
        msg = ("route didn't pass control communities checks - "
               "NOT ANNOUNCING {} TO {{inst}}".format(pref))
        self.log_contains(self.rs, msg, {"inst": self.AS1})

    def test_040_custom_bgp_community_std(self):
        """{}: custom BGP community (std)"""
        for inst in (self.AS2, self.AS131073):
            self.receive_route(inst, self.DATA["AS1_good1"], self.rs,
                               std_comms=["65501:65501"])

    def test_040_custom_bgp_community_ext(self):
        """{}: custom BGP community (ext)"""
        for inst in (self.AS2, self.AS131073):
            self.receive_route(inst, self.DATA["AS1_good1"], self.rs,
                               ext_comms=["rt:65501:65501"])

    def test_040_custom_bgp_community_lrg(self):
        """{}: custom BGP community (lrg)"""
        for inst in (self.AS2, self.AS131073):
            self.receive_route(inst, self.DATA["AS1_good1"], self.rs,
                               lrg_comms=["999:65501:65501"])

    def test_041_custom_bgp_community_scrubbed_inbound(self):
        """{}: custom BGP community scrubbed"""
        self.receive_route(self.rs, self.DATA["AS2_bad_cust_comm1"], self.AS2,
                           std_comms=[])
        for inst in (self.AS1, self.AS131073):
            self.receive_route(inst, self.DATA["AS2_bad_cust_comm1"], self.rs,
                               std_comms=[])

    def test_900_reconfigure(self):
        """{}: reconfigure"""
        self.rs.reload_config()
        self.test_020_sessions_up()

class BGPCommunitiesScenarioBIRD(BGPCommunitiesScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder

    TARGET_VERSION = None

    IP_VER = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER,
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bird/bird.conf"
                )
            ]
        )

class BGPCommunitiesScenarioBIRD2(BGPCommunitiesScenarioBIRD):
    __test__ = False

    TARGET_VERSION = "2.0.7"

class BGPCommunitiesScenarioOpenBGPD(LiveScenario_TagRejectPolicy,
                                     BGPCommunitiesScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder

    TARGET_VERSION = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bgpd.conf"
                )
            ]
        )


class BGPCommunitiesScenarioOpenBGPDPrevious(BGPCommunitiesScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDPreviousInstance.TARGET_VERSION


class BGPCommunitiesScenarioOpenBGPDLatest(BGPCommunitiesScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDLatestInstance.TARGET_VERSION



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from six.moves import queue
import time
import threading

from ..errors import BuilderError, ARouteServerError

class BaseConfigEnricherThread(threading.Thread):

    DESCR = None

    def __init__(self, tasks_q, errors_q, lock):
        threading.Thread.__init__(self)

        self.tasks_q = tasks_q
        self.errors_q = errors_q
        self.lock = lock

    def do_task(self, task):
        raise NotImplementedError()

    def save_data(self, task, data):
        raise NotImplementedError()

    def run(self):
        logging.debug("{} thread {} started".format(self.DESCR, self.name))

        while True:
            try:
                task = self.tasks_q.get(block=True, timeout=0.1)
            except queue.Empty:
                break

            try:
                data = self.do_task(task)
                if data:
                    with self.lock:
                        self.save_data(task, data)
            except Exception as e:
                if isinstance(e, ARouteServerError):
                    if str(e):
                        logging.error(
                            "{} thread {} error: {}".format(
                                self.DESCR, self.name,
                                str(e)
                            )
                        )
                else:
                    logging.error(
                        "{} thread {} unhandled exception: {}".format(
                            self.DESCR, self.name,
                            str(e) if str(e) else "error unknown"
                        ),
                        exc_info=True
                    )

                try:
                    self.errors_q.put_nowait(True)
                except queue.Full:
                    pass

            self.tasks_q.task_done()

        logging.debug("{} thread {} stopped".format(
            self.DESCR, self.name))

class QueueLengthMonitor(threading.Thread):

    def __init__(self, tasks_q, descr, threads):
        threading.Thread.__init__(self)

        self.tasks_q = tasks_q
        self.descr = descr
        self.threads = threads

        self.name = "'{}' enricher queue monitor".format(self.descr)
        self.daemon = True

        self.done = False

    def run(self):
        start_time = int(time.time())
        while not self.done and self.tasks_q.qsize() > 0:
            time.sleep(1)

            # Print n. of remaining tasks every 10 seconds
            if (int(time.time()) - start_time) % 10 == 0:
                tasks_left = self.tasks_q.qsize()
                logging.info("Enricher '{}', {} tasks left".format(
                    self.descr, tasks_left
                ))

class BaseConfigEnricher(object):

    WORKER_THREAD_CLASS = None

    def __init__(self, builder, threads):
        self.builder = builder
        self.threads = threads
        self.tasks_q = queue.Queue()
        self.errors_q = queue.Queue(maxsize=1)

    def prepare(self):
        pass

    def _config_thread(self, thread):
        pass

    def add_tasks(self):
        raise NotImplementedError()

    def enrich(self):
        logging.info(
            "Enricher '{}' started".format(self.WORKER_THREAD_CLASS.DESCR)
        )
        start_time = int(time.time())

        self.prepare()

        lock = threading.Lock()

        threads = []
        for i in range(self.threads):
            t = self.WORKER_THREAD_CLASS(
                self.tasks_q, self.errors_q, lock
            )
            self._config_thread(t)
            threads.append(t)

        self.add_tasks()

        for t in threads:
            t.start()

        q_monitor = QueueLengthMonitor(self.tasks_q,
                                       self.WORKER_THREAD_CLASS.DESCR,
                                       self.threads)
        q_monitor.start()

        self.tasks_q.join()

        q_monitor.done = True
        q_monitor.join(timeout=5)

        stop_time = int(time.time())

        try:
            self.errors_q.get_nowait()
            logging.error(
                "Enricher '{}' completed with errors after {} seconds".format(
                    self.WORKER_THREAD_CLASS.DESCR, stop_time - start_time
                )
            )
            raise BuilderError()
        except queue.Empty:
            logging.info(
                "Enricher '{}' completed successfully after {} seconds".format(
                    self.WORKER_THREAD_CLASS.DESCR, stop_time - start_time
                )
            )
            return



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPDInstance, \
                                                          OpenBGPDPreviousInstance, \
                                                          OpenBGPDLatestInstance

class TagASSetScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance(),

            cls.CLIENT_INSTANCE_CLASS(
                "AS1",
                cls.DATA["AS1_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS4",
                cls.DATA["AS4_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS4.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS5",
                cls.DATA["AS5_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS5.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS6",
                cls.DATA["AS6_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS6.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
        ]

    def set_instance_variables(self):
        self.rs = self._get_instance_by_name("rs")
        self.AS1 = self._get_instance_by_name("AS1")
        self.AS2 = self._get_instance_by_name("AS2")
        self.AS4 = self._get_instance_by_name("AS4")
        self.AS5 = self._get_instance_by_name("AS5")
        self.AS6 = self._get_instance_by_name("AS6")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1)
        self.session_is_up(self.rs, self.AS2)
        self.session_is_up(self.rs, self.AS4)
        self.session_is_up(self.rs, self.AS5)
        self.session_is_up(self.rs, self.AS6)

    def test_060_AS2_whitelist_wl_wl(self):
        """{}: AS2 white list, prefix WL, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_wl_origin_wl"],
                           self.AS2, as_path="2 21", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS2_whitelist_wl_ko(self):
        """{}: AS2 white list, prefix WL, origin ko"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS2_pref_wl_origin_ko"],
                           self.AS2, as_path="2 3", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS2_whitelist_ko_wl(self):
        """{}: AS2 white list, prefix ko, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_ko_origin_wl"],
                           self.AS2, as_path="2 21", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS4_whitelist_wl_wl(self):
        """{}: AS4 white list, prefix WL, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_wl_origin_wl"],
                           self.AS4, as_path="4 41", next_hop=self.AS4,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS4_whitelist_ko_wl(self):
        """{}: AS4 white list, prefix ko, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_ko_origin_wl"],
                           self.AS4, as_path="4 41", next_hop=self.AS4,
                           std_comms=["999:64513", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS4_whitelist_wl_ko(self):
        """{}: AS4 white list, prefix WL, origin ko"""
        self.receive_route(self.rs, self.DATA["AS4_pref_wl_origin_ko"],
                           self.AS4, as_path="4 3", next_hop=self.AS4,
                           ext_comms=[],
                           filtered=True)

    def test_060_AS4_route_whitelist_1(self):
        """{}: AS4 route white list, ok (exact)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515", "999:0:64517"])
        self.receive_route(self.rs, self.DATA["AS4_routewl_1"],
                           self.AS4, as_path="4 44", next_hop=self.AS4,
                           std_comms=["999:64513", "999:64515", "999:64517"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS4_route_whitelist_2(self):
        """{}: AS4 route white list, reject (more spec)"""
        self.receive_route(self.rs, self.DATA["AS4_routewl_2"],
                           self.AS4, as_path="4 44", next_hop=self.AS4,
                           ext_comms=[],
                           filtered=True)

    def test_060_AS4_route_whitelist_3(self):
        """{}: AS4 route white list, ok (more spec)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515", "999:0:64517"])
        self.receive_route(self.rs, self.DATA["AS4_routewl_3"],
                           self.AS4, as_path="4 43", next_hop=self.AS4,
                           std_comms=["999:64513", "999:64515", "999:64517"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS4_route_whitelist_4(self):
        """{}: AS4 route white list, reject (origin KO)"""
        self.receive_route(self.rs, self.DATA["AS4_routewl_4"],
                           self.AS4, as_path="4 45", next_hop=self.AS4,
                           ext_comms=[],
                           filtered=True)

    def test_060_AS4_route_whitelist_5(self):
        """{}: AS4 route white list, ok (origin any)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515", "999:0:64517"])
        self.receive_route(self.rs, self.DATA["AS4_routewl_5"],
                           self.AS4, as_path="4 45", next_hop=self.AS4,
                           std_comms=["999:64513", "999:64515", "999:64517"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS5_whitelist_wl_ko(self):
        """{}: AS5 white list, prefix WL, origin ko"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS5_pref_wl_origin_ko"],
                           self.AS5, as_path="5 3", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS5_whitelist_wl_wl(self):
        """{}: AS5 white list, prefix WL, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS5_pref_wl_origin_wl"],
                           self.AS5, as_path="5 51", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS5_whitelist_ko_wl(self):
        """{}: AS5 white list, prefix ko, origin WL"""
        self.receive_route(self.rs, self.DATA["AS5_pref_ko_origin_wl"],
                           self.AS5, as_path="5 51", next_hop=self.AS5,
                           ext_comms=[],
                           filtered=True)

    def test_070_AS6_roas_as_route_objects_1(self):
        """{}: AS6 RPKI ROAs as route objects: invalid origin ASN"""
        self.receive_route(self.rs, self.DATA["AS2_roa1"],
                           self.AS6, as_path="6 2", next_hop=self.AS6,
                           ext_comms=[],
                           filtered=True, reject_reason=9)

    def test_900_reconfigure(self):
        """{}: reconfigure"""
        self.rs.reload_config()
        self.test_020_sessions_up()

class TagASSetScenario_WithAS_SETs(object):

    AS_SET = {
        "AS1": [1],
        "AS-AS2": [2],
        "AS-AS4": [4],
        "AS-AS5_FROM_PDB": [5],
        "AS6": [6, 3]
    }
    R_SET = {
        "AS1": [
            "AS1_allowed_prefixes"
        ],
        "AS-AS2": [
            "AS2_allowed_prefixes"
        ],
        "AS-AS4": [
            "AS4_allowed_prefixes"
        ],
        "AS-AS5_FROM_PDB": [
            "AS5_allowed_prefixes"
        ],
        "AS6": [
            "AS6_allowed_prefixes"
        ]
    }

    def _set_lrg_comms(self, lst):
        return lst

    def test_030_AS2_prefix_ok_origin_ok(self):
        """{}: AS2 prefix ok origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_ok_origin_ok1"], self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_030_AS2_prefix_ko_origin_ok(self):
        """{}: AS2 prefix ko origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_ko_origin_ok1"], self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_030_AS2_prefix_ok_origin_ko(self):
        """{}: AS2 prefix ok origin ko"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS3_pref_ok_origin_ko1"], self.AS2, as_path="2 3", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_030_AS2_prefix_ko_origin_ko(self):
        """{}: AS2 prefix ko origin ko"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS3_pref_ko_origin_ko1"], self.AS2, as_path="2 3", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_040_AS4_prefix_ok_origin_ok(self):
        """{}: AS4 prefix ok origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_ok_origin_ok1"], self.AS4, as_path="4", next_hop=self.AS4,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_040_AS4_prefix_ko_origin_ok(self):
        """{}: AS4 prefix ko origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_ko_origin_ok1"], self.AS4, as_path="4", next_hop=self.AS4,
                           std_comms=["999:64513", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_040_AS4_origin_filtered(self):
        """{}: AS4 route filtered (origin ko)"""
        self.receive_route(self.rs, self.DATA["AS3_pref_ok_origin_ko2"],
                           self.AS4, as_path="4 3", next_hop=self.AS4,
                           ext_comms=[],
                           filtered=True, reject_reason=9)

    def test_040_AS4_prefix_origin_filtered(self):
        """{}: AS4 route filtered (prefix ko, origin ko)"""
        self.receive_route(self.rs, self.DATA["AS3_pref_ko_origin_ko1"],
                           self.AS4, as_path="4 3", next_hop=self.AS4,
                           ext_comms=[],
                           filtered=True, reject_reason=9)

    def test_050_AS5_prefix_ok_origin_ok(self):
        """{}: AS5 prefix ok origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS5_pref_ok_origin_ok1"], self.AS5, as_path="5", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_050_AS5_prefix_filtered(self):
        """{}: AS5 route filtered (prefix ko)"""
        self.receive_route(self.rs, self.DATA["AS5_pref_ko_origin_ok1"],
                           self.AS5, as_path="5", next_hop=self.AS5,
                           ext_comms=[],
                           filtered=True, reject_reason=12)

    def test_050_AS5_prefix_ok_origin_ko(self):
        """{}: AS5 prefix ok origin ko"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS3_pref_ok_origin_ko3"], self.AS5, as_path="5 3", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_050_AS5_origin_filtered(self):
        """{}: AS5 route filtered (prefix ko, origin ko)"""
        # AS5 is configured with prefix enforcement only, that's why
        # the reject reason should be 12 (prefix not in IRRDBs) and
        # not 9 (origin ASN not in IRRDBs).
        self.receive_route(self.rs, self.DATA["AS3_pref_ko_origin_ko1"],
                           self.AS5, as_path="5 3", next_hop=self.AS5,
                           ext_comms=[],
                           filtered=True, reject_reason=12)

    def test_060_AS2_whitelist_wl_ok(self):
        """{}: AS2 white list, prefix WL, origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_wl_origin_ok"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS2_whitelist_ok_wl(self):
        """{}: AS2 white list, prefix ok, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_ok_origin_wl"],
                           self.AS2, as_path="2 21", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS4_whitelist_wl_ok(self):
        """{}: AS4 white list, prefix WL, origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_wl_origin_ok"],
                           self.AS4, as_path="4", next_hop=self.AS4,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS4_whitelist_ok_wl(self):
        """{}: AS4 white list, prefix ok, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_ok_origin_wl"],
                           self.AS4, as_path="4 41", next_hop=self.AS4,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS5_whitelist_wl_ok(self):
        """{}: AS5 white list, prefix WL, origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS5_pref_wl_origin_ok"],
                           self.AS5, as_path="5", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS5_whitelist_ok_wl(self):
        """{}: AS5 white list, prefix ok, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS5_pref_ok_origin_wl"],
                           self.AS5, as_path="5 51", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_070_AS6_roas_as_route_objects_2(self):
        """{}: AS6 RPKI ROAs as route objects: ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514", "999:0:64516"])
        self.receive_route(self.rs, self.DATA["AS3_roa2"],
                           self.AS6, as_path="6 3", next_hop=self.AS6,
                           std_comms=["999:64513", "999:64514", "999:64516"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_070_AS2_roas_as_route_objects_tag_only(self):
        """{}: AS2 RPKI ROAs as route objects: tag only (w/ prefix_validated_via_rpki_roas)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514", "999:0:64516"])
        self.receive_route(self.rs, self.DATA["AS2_roa2"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64514", "999:64516"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_080_AS6_arin_whois_db_1(self):
        """{}: AS6 ARIN Whois DB: ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514", "999:0:64518"])
        self.receive_route(self.rs, self.DATA["AS3_arin1"],
                           self.AS6, as_path="6 3", next_hop=self.AS6,
                           std_comms=["999:64513", "999:64514", "999:64518"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_080_AS2_arin_whois_db_tag_only(self):
        """{}: AS2 ARIN Whois DB: tag only (w/ prefix_validated_via_arin_whois_db_dump)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514", "999:0:64518"])
        self.receive_route(self.rs, self.DATA["AS2_arin1"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64514", "999:64518"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS2_roa_and_arin_tag_only(self):
        """{}: AS2 ROA + ARIN Whois DB: tag only (w/ comms [arin_whois_db_dump, rpki_roas])"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514", "999:0:64516", "999:0:64518"])
        self.receive_route(self.rs, self.DATA["AS2_roa3_arin2"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64514", "999:64516", "999:64518"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS6_roa_and_arin_enforced(self):
        """{}: AS6 ROA + ARIN Whois DB: enforce (w/ comms [arin_whois_db_dump, rpki_roas])"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514", "999:0:64516", "999:0:64518"])
        self.receive_route(self.rs, self.DATA["AS3_roa3_arin2"],
                           self.AS6, as_path="6 3", next_hop=self.AS6,
                           std_comms=["999:64513", "999:64514", "999:64516", "999:64518"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS2_pref_ok_orig_ok_roa_tag_only(self):
        """{}: AS2 prefix ok, origin ok, ROA: tag only (w/ prefix_validated_via_rpki_roas)"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514", "999:0:64516"])
        self.receive_route(self.rs, self.DATA["AS2_ok_ok_roa3"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64514", "999:64516"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS2_pref_ok_orig_ok_arin_tag_only(self):
        """{}: AS2 prefix ok, origin ok, ARIN: tag only (w/ prefix_validated_via_arin_whois_db_dump)"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514", "999:0:64518"])
        self.receive_route(self.rs, self.DATA["AS2_ok_ok_arin3"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64514", "999:64518"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS6_pref_ok_orig_ok_roa_arin_enforced(self):
        """{}: AS6 prefix ok, origin ok, ROA + ARIN: enforce (w/ comms [arin_whois_db_dump, rpki_roas])"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64514", "999:0:64516", "999:0:64518"])
        self.receive_route(self.rs, self.DATA["AS6_ok_ok_roa6_arin6"],
                           self.AS6, as_path="6", next_hop=self.AS6,
                           std_comms=["999:64512", "999:64514", "999:64516", "999:64518"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

class TagASSetScenario_EmptyAS_SETs(object):

    AS_SET = {
        "AS1": [],
        "AS-AS2": [],
        "AS-AS4": [],
        "AS-AS5_FROM_PDB": [],
        "AS6": []
    }
    R_SET = {
        "AS1": [
        ],
        "AS-AS2": [
        ],
        "AS-AS4": [
        ],
        "AS-AS5_FROM_PDB": [
        ],
        "AS6": [
        ]
    }

    def _set_lrg_comms(self, lst):
        return lst

    def test_030_AS2_no_enforcement(self):
        """{}: AS2 no enforcement, prefix and origin not in AS-SET"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        for pref in (self.DATA["AS2_pref_ok_origin_ok1"],
                     self.DATA["AS2_pref_ko_origin_ok1"],
                     self.DATA["AS3_pref_ok_origin_ko1"],
                     self.DATA["AS3_pref_ko_origin_ko1"]):
            self.receive_route(self.rs, pref, self.AS2, next_hop=self.AS2,
                            std_comms=["999:64513", "999:64515"],
                            ext_comms=[],
                            lrg_comms=lrg_comms)

    def test_040_AS4_origin_enforcement(self):
        """{}: AS4 origin enforcement"""
        for pref in (self.DATA["AS4_pref_ok_origin_ok1"],
                     self.DATA["AS4_pref_ko_origin_ok1"],
                     self.DATA["AS3_pref_ok_origin_ko2"],
                     self.DATA["AS3_pref_ko_origin_ko1"]):
            self.receive_route(self.rs, pref, self.AS4, next_hop=self.AS4,
                               ext_comms=[],
                               filtered=True, reject_reason=(9, 12))

    def test_050_AS4_prefix_enforcement(self):
        """{}: AS4 prefix enforcement"""
        for pref in (self.DATA["AS5_pref_ok_origin_ok1"],
                     self.DATA["AS5_pref_ko_origin_ok1"],
                     self.DATA["AS3_pref_ok_origin_ko3"],
                     self.DATA["AS3_pref_ko_origin_ko1"]):
            self.receive_route(self.rs, pref, self.AS5, next_hop=self.AS5,
                               ext_comms=[],
                               filtered=True, reject_reason=(9, 12))

    def test_060_AS2_whitelist_wl_ok(self):
        """{}: AS2 white list, prefix WL, origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS2_pref_wl_origin_ok"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64512", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS2_whitelist_ok_wl(self):
        """{}: AS2 white list, prefix ok, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS2_pref_ok_origin_wl"],
                           self.AS2, as_path="2 21", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS4_whitelist_wl_ok(self):
        """{}: AS4 white list, prefix WL, origin ok"""
        self.receive_route(self.rs, self.DATA["AS4_pref_wl_origin_ok"],
                           self.AS4, as_path="4", next_hop=self.AS4,
                           ext_comms=[],
                           filtered=True)

    def test_060_AS4_whitelist_ok_wl(self):
        """{}: AS4 white list, prefix ok, origin WL"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64514"])
        self.receive_route(self.rs, self.DATA["AS4_pref_ok_origin_wl"],
                           self.AS4, as_path="4 41", next_hop=self.AS4,
                           std_comms=["999:64513", "999:64514"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS5_whitelist_wl_ok(self):
        """{}: AS5 white list, prefix WL, origin ok"""
        lrg_comms = self._set_lrg_comms(["999:0:64512", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS5_pref_wl_origin_ok"],
                           self.AS5, as_path="5", next_hop=self.AS5,
                           std_comms=["999:64512", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_060_AS5_whitelist_ok_wl(self):
        """{}: AS5 white list, prefix ok, origin WL"""
        self.receive_route(self.rs, self.DATA["AS5_pref_ok_origin_wl"],
                           self.AS5, as_path="5 51", next_hop=self.AS5,
                           ext_comms=[],
                           filtered=True)

    def test_070_AS6_roas_as_route_objects_2(self):
        """{}: AS6 RPKI ROAs as route objects: ko"""
        self.receive_route(self.rs, self.DATA["AS3_roa2"],
                           self.AS6, as_path="6 3", next_hop=self.AS6,
                           ext_comms=[],
                           filtered=True)

    def test_070_AS2_roas_as_route_objects_tag_only(self):
        """{}: AS2 RPKI ROAs as route objects: tag only (w/o prefix_validated_via_rpki_roas)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS2_roa2"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_080_AS6_arin_whois_db_1(self):
        """{}: AS6 ARIN Whois DB: ok (solely because of route white list)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515", "999:0:64517"])
        self.receive_route(self.rs, self.DATA["AS3_arin1"],
                           self.AS6, as_path="6 3", next_hop=self.AS6,
                           std_comms=["999:64513", "999:64515", "999:64517"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_080_AS2_arin_whois_db_tag_only(self):
        """{}: AS2 ARIN Whois DB: tag only (w/o prefix_validated_via_arin_whois_db_dump)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS2_arin1"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS2_roa_and_arin_tag_only(self):
        """{}: AS2 ROA + ARIN Whois DB: tag only (w/o comms [arin_whois_db_dump, rpki_roas])"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS2_roa3_arin2"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS6_roa_and_arin_enforced(self):
        """{}: AS6 ROA + ARIN Whois DB: enforced (rejected)"""
        self.receive_route(self.rs, self.DATA["AS3_roa3_arin2"],
                           self.AS6, as_path="6 3", next_hop=self.AS6,
                           ext_comms=[],
                           filtered=True)

    def test_090_AS2_pref_ok_orig_ok_roa_tag_only(self):
        """{}: AS2 prefix ok, origin ok, ROA: tag only (w/o prefix_validated_via_rpki_roas)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS2_ok_ok_roa3"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS2_pref_ok_orig_ok_arin_tag_only(self):
        """{}: AS2 prefix ok, origin ok, ARIN: tag only (w/o prefix_validated_via_arin_whois_db_dump)"""
        lrg_comms = self._set_lrg_comms(["999:0:64513", "999:0:64515"])
        self.receive_route(self.rs, self.DATA["AS2_ok_ok_arin3"],
                           self.AS2, as_path="2", next_hop=self.AS2,
                           std_comms=["999:64513", "999:64515"],
                           ext_comms=[],
                           lrg_comms=lrg_comms)

    def test_090_AS6_pref_ok_orig_ok_roa_arin_enforced(self):
        """{}: AS6 prefix ok, origin ok, ROA + ARIN: rejected"""
        self.receive_route(self.rs, self.DATA["AS6_ok_ok_roa6_arin6"],
                           self.AS6, as_path="6", next_hop=self.AS6,
                           ext_comms=[],
                           filtered=True)

class TagASSetScenarioBIRD(TagASSetScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder
    TARGET_VERSION = None
    IP_VER = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER,
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bird/bird.conf"
                )
            ]
        )

class TagASSetScenarioBIRD2(TagASSetScenarioBIRD):
    __test__ = False

    TARGET_VERSION = "2.0.7"

class TagASSetScenarioOpenBGPD(LiveScenario_TagRejectPolicy, TagASSetScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder

    TARGET_VERSION = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bgpd.conf"
                )
            ]
        )

class TagASSetScenarioOpenBGPDPrevious(TagASSetScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDPreviousInstance.BGP_SPEAKER_VERSION

class TagASSetScenarioOpenBGPDLatest(TagASSetScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDLatestInstance.BGP_SPEAKER_VERSION



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re

from ..errors import ConfigError
from ..ipaddresses import IPAddress, IPNetwork


class ConfigParserValidator(object):

    def __init__(self, **kwargs):
        self.mandatory = True
        if "mandatory" in kwargs:
            self.mandatory = bool(kwargs["mandatory"])

        self.default = None
        if "default" in kwargs:
            self.default = kwargs["default"]

    def _validate(self, v):
        raise NotImplementedError()

    def validate(self, v):
        try:
            if v is None:
                if self.mandatory:
                    if self.default is not None:
                        return self._validate(self.default)
                    raise ConfigError()
                else:
                    return None

            if isinstance(v, str):
                if v.strip() == "":
                    if self.mandatory:
                        if self.default is not None:
                            return self._validate(self.default)
                        raise ConfigError()
                    else:
                        return None

        except ConfigError:
            raise ConfigError("Can't be empty")

        return self._validate(v)

class ValidatorUInt(ConfigParserValidator):

    def _validate(self, v):
        if isinstance(v, str):
            if v.strip().isdigit():
                return int(v)
            else:
                raise ConfigError()
        if isinstance(v, int):
            if v >= 0:
                return v
            else:
                raise ConfigError()
        raise ConfigError()

class ValidatorText(ConfigParserValidator):

    def _validate(self, v):
        if v is not None:
            return str(v)
        return None

class ValidatorASSet(ValidatorText):

    def _validate(self, v):
        source_macro = re.match("^([A-Za-z]+)::(.+)$", v.strip())
        if source_macro:
            source = source_macro.group(1)
            macro = source_macro.group(2)
        else:
            source = None
            macro = v.strip()

        try:
            macro = self._parse_asset(macro)
        except ConfigError as e:
            msg = "Invalid AS-SET: {}".format(v)
            if str(e):
                msg += ", " + str(e)
            raise ConfigError(msg)

        if source:
            return "{}::{}".format(source, macro)
        return macro

    def _parse_asset(self, v):
        macro = v.strip()

        # "Many objects in RPSL have a name.  An <object-name> is
        # made up of letters, digits, the character underscore "_",
        # and the character hyphen "-"; the first character of a
        # name must be a letter, and the last character of a name
        # must be a letter or a digit.
        # An AS number x is represented as the string "ASx".  That
        # is, the AS 226 is represented as AS226."
        # https://datatracker.ietf.org/doc/html/rfc2622#section-2
        #
        # "A hierarchical set name is a sequence of set names and
        # AS numbers separated by colons ":".
        # At least one component of such a name must be an actual
        # set name (i.e. start with one of the prefixes above)."
        # https://datatracker.ietf.org/doc/html/rfc2622#section-5
        if re.match("^AS[\d]+$", macro):
            return macro

        as_dash_found = False
        parts = []
        for part in macro.split(":"):
            name = part.strip().upper()
            if not re.match("^(?:AS[\d]+|AS-[A-Z0-9_\-]*[A-Z0-9])$", name):
                raise ConfigError("invalid name {}".format(name))
            if name.startswith("AS-"):
                as_dash_found = True
            parts.append(name)
        macro = ":".join(parts)

        if not as_dash_found:
            raise ConfigError("no ""AS-"" found")

        if not macro:
            raise ConfigError()

        return macro

class ValidatorASN(ConfigParserValidator):

    def _validate(self, v):
        try:
            asn = ValidatorUInt().validate(v)
            if asn == 0:
                raise ConfigError()
            return asn
        except ConfigError:
            raise ConfigError("Invalid ASN: {}".format(v))

class ValidatorASNList(ConfigParserValidator):

    def _validate(self, v):
        if isinstance(v, str):
            parts = v.split(",")
        elif isinstance(v, list):
            parts = v
        elif isinstance(v, int):
            return [ValidatorASN().validate(v)]
        else:
            raise ConfigError(
                "Invalid format: must be a list or a "
                "comma-separated list of ASNs"
            )

        asns = []
        for part in parts:
            asns.append(ValidatorASN().validate(part))
        return asns

class ValidatorIPAddr(ConfigParserValidator):

    def _validate(self, v):
        try:
            ip = IPAddress(v)
            return ip.ip
        except:
            raise ConfigError("Invalid IP address: {}".format(v))

class ValidatorIPv4Addr(ConfigParserValidator):

    def _validate(self, v):
        try:
            ip = IPAddress(v)
            if not ip.version == 4:
                raise ValueError()
            return ip.ip
        except:
            raise ConfigError("Invalid IPv4 address: {}".format(v))

class ValidatorIPv6Addr(ConfigParserValidator):

    def _validate(self, v):
        try:
            ip = IPAddress(v)
            if not ip.version == 6:
                raise ValueError()
            return ip.ip
        except:
            raise ConfigError("Invalid IPv6 address: {}".format(v))

class ValidatorListOf(ConfigParserValidator):

    def __init__(self, cls, *args, **kwargs):
        ConfigParserValidator.__init__(self, *args, **kwargs)
        self.cls = cls

    def _validate(self, l):
        if not isinstance(l, list):
            raise ConfigError("Invalid format: must be a list")

        for v in l:
            validator = self.cls()
            v = validator.validate(v)

        return l

class ValidatorPrefixListEntry(ConfigParserValidator):

    def _validate(self, v):
        if not isinstance(v, dict):
            raise ConfigError("Invalid prefix list entry format: must be dict")

        for prop in v:
            if prop not in ("prefix", "length", "comment", "exact", "ge", "le",
                            "max_length"):
                raise ConfigError(
                    "Unknown statement '{}' in prefix list entry "
                    "definition".format(prop)
                )
        for prop in ("prefix", "length"):
            if prop not in v:
                raise ConfigError(
                    "Missing '{}' in prefix list entry".format(prop)
                )

        try:
            ip_obj = IPNetwork(v["prefix"])
        except:
            raise ConfigError("Invalid prefix ID: {}".format(v["prefix"]))

        try:
            pref_len = ValidatorUInt().validate(v["length"])

            if pref_len < 0:
                raise ConfigError()
            if ip_obj.version == 4:
                if pref_len > 32:
                    raise ConfigError()
            if ip_obj.version == 6:
                if pref_len > 128:
                    raise ConfigError()
        except:
            raise ConfigError("Invalid prefix length: {}".format(v["length"]))

        v["prefix"] = ip_obj.ip
        v["length"] = pref_len
        v["comment"] = str(v["comment"]) if "comment" in v else None
        v["max_length"] = ip_obj.max_prefixlen

        if "exact" in v:
            try:
                v["exact"] = ValidatorBool().validate(v["exact"])
            except:
                raise ConfigError(
                    "Invalid 'exact' flag value: {}".format(v["exact"])
                )
        else:
            v["exact"] = False

        def _validate_ge_le(ge_or_le, v, pref_len, ip_obj):
            try:
                v[ge_or_le] = ValidatorUInt().validate(v[ge_or_le])
            except:
                raise ConfigError("Invalid '{}' value ({}): "
                    "must be a positive integer".format(
                    ge_or_le, v[ge_or_le]
                ))

            if v[ge_or_le] < pref_len:
                raise ConfigError(
                    "'{}' ({}) must be greater than or equal to "
                    "the prefix-len ({})".format(
                        ge_or_le, v[ge_or_le], pref_len
                    )
                )
            if v[ge_or_le] > ip_obj.max_prefixlen:
                raise ConfigError(
                    "'{}' ({}) must be less than or equal to "
                    "the max prefix-len ({})".format(
                        ge_or_le, v[ge_or_le], ip_obj.max_prefixlen
                    )
                )

        if "ge" in v and v["ge"]:
            _validate_ge_le("ge", v, pref_len, ip_obj)
        else:
            v["ge"] = None

        if "le" in v and v["le"]:
            _validate_ge_le("le", v, pref_len, ip_obj)
        else:
            v["le"] = None

        if v["ge"] and v["le"]:
            if v["ge"] > v["le"]:
                raise ConfigError(
                    "'ge' must be less than or equal to 'le'"
                )

        if v["ge"] or v["le"]:
            if v["exact"]:
                raise ConfigError(
                    "Can't set 'ge' and 'le' when 'exact' is True"
                )

        return v

class ValidatorWhiteListRouteEntry(ConfigParserValidator):

    def _validate(self, v):
        # Bad trick here: remove 'asn' in order to have
        # the rest of the dict validated as if it was a
        # standard prefix-list entry.
        # Then, add it back again.
        asn = None
        if "asn" in v:
            v["asn"] = ValidatorASN(mandatory=False).validate(v["asn"])
            asn = v["asn"]
            del v["asn"]

        ValidatorPrefixListEntry()._validate(v)

        v["asn"] = asn

        return v

class ValidatorBool(ConfigParserValidator):

    def _validate(self, v):
        try:
            if isinstance(v, bool):
                return v
            if isinstance(v, int):
                if v == 0:
                    return False
                elif v == 1:
                    return True
                else:
                    raise ConfigError()
            elif v.lower() in ["true", "yes", "t", "1"]:
                return True
            elif v.lower() in ["false", "no", "f", "0"]:
                return False
            else:
                raise ConfigError()
        except:
            raise ConfigError("Invalid boolean value: {}".format(v))

class ValidatorOption(ConfigParserValidator):

    def __init__(self, name, options, **kwargs):
        ConfigParserValidator.__init__(self, **kwargs)
        self.name = name
        self.options = options

    def _validate(self, v):
        if v is None and None in self.options:
            return v
        if v in self.options:
            return v
        raise ConfigError(
            "Invalid option for '{}': '{}'; "
            "it must be one of {}".format(
                self.name, v,
                ", ".join(["null" if o is None else "'{}'".format(o)
                           for o in self.options])
            )
        )

class ValidatorIPMinMaxLen(ConfigParserValidator):

    def __init__(self, ver, **kwargs):
        ConfigParserValidator.__init__(self, **kwargs)
        self.ver = ver

    def _validate(self, v):
        if not isinstance(v, dict):
            raise ConfigError(
                "Invalid format for IPv{} min/max length".format(self.ver)
            )

        if self.ver == 4:
            max_val = 32
        else:
            max_val = 128

        for min_max in ("min", "max"):
            if min_max not in v:
                raise ConfigError(
                    "Missing '{}' in the IPv{} min/max length".format(
                        min_max, self.ver
                    )
                )
            try:
                val = ValidatorUInt().validate(v[min_max])
            except ConfigError:
                raise ConfigError(
                    "Invalid '{}' in the IPv{} min/max length: {}".format(
                        min_max, self.ver, v[min_max]
                    )
                )

            if val > max_val:
                raise ConfigError(
                    "Value of '{}' in the IPv{} min/max length out of "
                    "range; given {}, allowed: {}-{}".format(
                        min_max, self.ver, val, 0, max_val
                    )
                )

        if int(v["min"]) > int(v["max"]):
            raise ConfigError(
                "In the IPv{} min/max length, the value of 'min' must be "
                "<= the value of 'max'".format(self.ver)
            )
        return {
            "min": int(v["min"]),
            "max": int(v["max"])
        }

class ValidatorMaxASPathLen(ConfigParserValidator):

    def _validate(self, v):
        try:
            val = ValidatorUInt().validate(v)
            if val >= 1 and val <= 64:
                return int(v)
            else:
                raise ConfigError()
        except ConfigError:
            raise ConfigError(
                "Invalid max_as_path_len: must be an integer "
                "between 1 and 64"
            )

class ValidatorCommunity(ConfigParserValidator):

    EXPECTED_PARTS_CNT = None

    def __init__(self, rs_as, **kwargs):
        ConfigParserValidator.__init__(self, **kwargs)
        self.rs_as = rs_as
        self.peer_as_macro_needed = kwargs.get("peer_as_macro_needed", False)
        self.dyn_val_macro_needed = kwargs.get("dyn_val_macro_needed", False)

    def _expand_rs_as_macro(self, v):
        if "rs_as" in v:
            if self.rs_as:
                return v.replace("rs_as", str(self.rs_as))
            else:
                raise ConfigError(
                    "Can't expand 'rs_as' macro in {}: "
                    "'rs_as' unknown".format(v)
                )
        else:
            return v

    def _get_parts(self, val):
        parts = list(map(str.strip, val.split(":")))
        if len(parts) != self.EXPECTED_PARTS_CNT:
            raise ConfigError()

        def check_dynamic_part(macro_name, needed):
            # dynamic part macro usage checks
            macro_found = False
            for part_idx in range(len(parts)):
                if parts[part_idx] != macro_name:
                    continue
                if needed:
                    macro_found = True
                    if part_idx != len(parts) - 1:
                        # dynamic macros allowed only in the last part
                        raise ConfigError("'{}' macro can be used only "
                                          "in the last part of the value".format(
                                              macro_name))
                else:
                    raise ConfigError("'{}' macro not allowed".format(macro_name))
            if needed and not macro_found:
                raise ConfigError("'{}' macro is mandatory in this "
                                  "community".format(macro_name))

        check_dynamic_part("peer_as", self.peer_as_macro_needed)
        check_dynamic_part("dyn_val", self.dyn_val_macro_needed)

        return parts

class ValidatorCommunityStd(ValidatorCommunity):

    EXPECTED_PARTS_CNT = 2

    def _validate(self, v):
        if not isinstance(v, str):
            raise ConfigError("value must be a string")

        val = self._expand_rs_as_macro(v)

        try:
            validated_parts = []
            parts = self._get_parts(val)
            for part in parts:
                if part.strip() in ["peer_as", "dyn_val"]:
                    validated_parts.append(part.strip())
                    continue
                part_val = ValidatorUInt().validate(part)
                if part_val < 0 or part_val > 65535:
                    raise ConfigError()
                validated_parts.append(str(int(part_val)))
            if parts[0] == "65535":
                raise ConfigError(
                    "range 65535:x is reserved"
                )
            return ":".join(validated_parts)
        except ConfigError as e:
            raise ConfigError(
                "Invalid BGP standard community: {}{}; "
                "it must be in the x:x format, with x = a 16-bit "
                "unsigned integer; the 'rs_as' macro "
                "can be used to represent the route "
                "server's ASN provided "
                "that it is a 16-bit ASN".format(
                    v, " - {}".format(str(e)) if str(e) else ""
                )
            )

class ValidatorCommunityLrg(ValidatorCommunity):

    EXPECTED_PARTS_CNT = 3

    def _validate(self, v):
        if not isinstance(v, str):
            raise ConfigError("value must be a string")

        val = self._expand_rs_as_macro(v)

        try:
            validated_parts = []
            parts = self._get_parts(val)
            for part in parts:
                if part.strip() in ["peer_as", "dyn_val"]:
                    validated_parts.append(part.strip())
                    continue
                part_val = ValidatorUInt().validate(part)
                if part_val < 0 or part_val > 4294967295:
                    raise ConfigError()
                validated_parts.append(str(int(part_val)))
            return ":".join(validated_parts)
        except ConfigError as e:
            raise ConfigError(
                "Invalid BGP large community: {}{}; "
                "it must be in the x:x:x format, with x = a 32-bit "
                "unsigned integer; the 'rs_as' and 'peer-as' macros "
                "can be used to represent, respectively, the route "
                "server's ASN and the destination peer's ASN".format(
                    v, " - {}".format(str(e)) if str(e) else ""
                )
            )

class ValidatorCommunityExt(ValidatorCommunity):

    EXPECTED_PARTS_CNT = 3

    def _validate(self, v):
        if not isinstance(v, str):
            raise ConfigError("value must be a string")

        val = self._expand_rs_as_macro(v)

        # TODO: should be improved
        try:
            validated_parts = []
            parts = self._get_parts(val)
            if parts[0].strip().lower() not in ("rt", "ro"):
                raise ConfigError()
            validated_parts.append(parts[0].strip().lower())
            for part in parts[1:]:
                if part.strip() in ["peer_as", "dyn_val"]:
                    validated_parts.append(part.strip())
                    continue
                part_val = ValidatorUInt().validate(part)
                if part_val < 0 or part_val > 4294967295:
                    raise ConfigError()
                validated_parts.append(str(int(part_val)))
            return ":".join(validated_parts)
        except ConfigError as e:
            raise ConfigError(
                "Invalid BGP extended community: {}{}; "
                "it must be in the k:x:x format, with k one of "
                "'rt' or 'ro' and x = an unsigned integer; "
                "the 'rs_as' and 'peer-as' macros "
                "can be used to represent, respectively, the route "
                "server's ASN and the destination peer's ASN".format(
                    v, " - {}".format(str(e)) if str(e) else ""
                )
            )

class ValidatorRTTThresholds(ConfigParserValidator):

    def _validate(self, v):
        if isinstance(v, str):
            lst = v.split(",")
        elif isinstance(v, list):
            lst = v
            pass
        else:
            raise ConfigError(
                "Invalid type: {} - it must be a list of integers".format(
                    type(v)
                )
            )

        res = []
        for x in lst:
            try:
                rtt = ValidatorUInt().validate(x)
            except:
                raise ConfigError(
                    "RTT thresholds list items must be "
                    "positive integers: {}".format(x)
                )
            if not res:
                res.append(rtt)
                continue
            if rtt in res:
                raise ConfigError(
                    "Duplicate RTT value found: {}".format(rtt)
                )
            if rtt < res[-1]:
                raise ConfigError(
                    "RTT thresholds list items must be "
                    "provided in ascending order: {} < {}".format(
                        rtt, res[-1]
                    )
                )
            res.append(rtt)

        return res



# Copyright (C) 2017-2020 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import six

from pierky.arouteserver.builder import OpenBGPDConfigBuilder, BIRDConfigBuilder
from pierky.arouteserver.tests.live_tests.base import LiveScenario, \
                                                      LiveScenario_TagRejectPolicy
from pierky.arouteserver.tests.live_tests.bird import BIRDInstance
from pierky.arouteserver.tests.live_tests.openbgpd import OpenBGPDPreviousInstance, \
                                                          OpenBGPDLatestInstance

class MaxPrefixScenario(LiveScenario):
    __test__ = False

    MODULE_PATH = __file__
    RS_INSTANCE_CLASS = None
    CLIENT_INSTANCE_CLASS = None
    CONFIG_BUILDER_CLASS = None

    @classmethod
    def _setup_instances(cls):
        cls.INSTANCES = [
            cls._setup_rs_instance(),

            cls.CLIENT_INSTANCE_CLASS(
                "AS1",
                cls.DATA["AS1_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS1.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS2",
                cls.DATA["AS2_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS2.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS3",
                cls.DATA["AS3_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS3.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
            cls.CLIENT_INSTANCE_CLASS(
                "AS4",
                cls.DATA["AS4_1_IPAddress"],
                [
                    (
                        cls.build_other_cfg("AS4.j2"),
                        "/etc/bird/bird.conf"
                    )
                ],
            ),
        ]

    def set_instance_variables(self):
        self.rs = self._get_instance_by_name("rs")
        self.AS1 = self._get_instance_by_name("AS1")
        self.AS2 = self._get_instance_by_name("AS2")
        self.AS3 = self._get_instance_by_name("AS3")
        self.AS4 = self._get_instance_by_name("AS4")

    def test_010_setup(self):
        """{}: instances setup"""
        pass

    def test_900_reconfigure(self):
        """{}: reconfigure"""
        self.rs.reload_config()

class MaxPrefixScenarioBIRD(MaxPrefixScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = BIRDConfigBuilder
    TARGET_VERSION = None
    IP_VER = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("bird", "main.j2", "rs.conf", cls.IP_VER,
                                     cfg_general="general_bird.yml",
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bird/bird.conf"
                )
            ]
        )

    def test_020_sessions_up(self):
        """{}: sessions are up"""
        self.session_is_up(self.rs, self.AS1)
        self.session_is_up(self.rs, self.AS2)
        self.session_is_up(self.rs, self.AS3)

    def test_030_blocked_sessions(self):
        """{}: log is populated"""
        log_tpl = "Protocol {{inst}} hits route import limit ({limit}), action: block"

        log = log_tpl.format(limit=4)
        self.log_contains(self.rs, log, {"inst": self.AS1})

        log = log_tpl.format(limit=3)
        self.log_contains(self.rs, log, {"inst": self.AS2})

        log = log_tpl.format(limit=2)
        self.log_contains(self.rs, log, {"inst": self.AS3})

        log = log_tpl.format(limit=6)
        self.log_contains(self.rs, log, {"inst": self.AS4})

    def _get_routes_from(self, asn, include_filtered=False):
        routes = []
        for prefix_num in (1,2,3,4,5,6,7):
            if "AS{}_pref{}".format(asn, prefix_num) not in self.DATA:
                continue
            routes.extend(
                self.rs.get_routes(
                    self.DATA["AS{}_pref{}".format(asn, prefix_num)],
                    include_filtered=include_filtered
                )
            )
        return routes

    def test_030_count_received_prefixes_AS1(self):
        """{}: number of prefixes received by rs from AS1"""

        self.assertEqual(len(self._get_routes_from(1)), 4)
        self.assertEqual(len(self._get_routes_from(1, include_filtered=True)), 5)

    def test_031_count_received_prefixes_AS2(self):
        """{}: number of prefixes received by rs from AS2"""

        self.assertEqual(len(self._get_routes_from(2)), 3)
        self.assertEqual(len(self._get_routes_from(2, include_filtered=True)), 5)

    def test_032_count_received_prefixes_AS3(self):
        """{}: number of prefixes received by rs from AS3"""

        self.assertEqual(len(self._get_routes_from(3)), 2)
        self.assertEqual(len(self._get_routes_from(3, include_filtered=True)), 5)

    def test_032_count_received_prefixes_AS4(self):
        """{}: number of prefixes received by rs from AS4"""

        self.assertEqual(len(self._get_routes_from(4)), 6)
        self.assertEqual(len(self._get_routes_from(4, include_filtered=True)), 7)

    def test_900_reconfigure(self):
        """{}: reconfigure"""
        self.rs.reload_config()
        self.test_020_sessions_up()

class MaxPrefixScenarioBIRD2(MaxPrefixScenarioBIRD):
    __test__ = False

    TARGET_VERSION = "2.0.7"

class MaxPrefixScenarioOpenBGPD(LiveScenario_TagRejectPolicy, MaxPrefixScenario):
    __test__ = False

    CONFIG_BUILDER_CLASS = OpenBGPDConfigBuilder

    TARGET_VERSION = None

    @classmethod
    def _setup_rs_instance(cls):
        return cls.RS_INSTANCE_CLASS(
            "rs",
            cls.DATA["rs_IPAddress"],
            [
                (
                    cls.build_rs_cfg("openbgpd", "main.j2", "rs.conf", None,
                                     cfg_general=cls._get_cfg_general("general_openbgpd.yml"),
                                     target_version=cls.TARGET_VERSION),
                    "/etc/bgpd.conf"
                )
            ]
        )

    def test_020_sessions_down(self):
        """{}: sessions are down"""
        for inst in (self.AS1, self.AS2, self.AS3, self.AS4):
            with six.assertRaisesRegex(self,
                AssertionError, "is not up"
            ):
                self.session_is_up(self.rs, inst)

    def test_030_clients_receive_maxpref_not(self):
        """{}: clients log max-prefix notification"""
        for inst in (self.AS1, self.AS2, self.AS3, self.AS4):
            self.log_contains(inst, "the_rs: Received: Maximum number of prefixes reached")

class MaxPrefixScenarioOpenBGPDPrevious(MaxPrefixScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDPreviousInstance.BGP_SPEAKER_VERSION

class MaxPrefixScenarioOpenBGPDLatest(MaxPrefixScenarioOpenBGPD):
    __test__ = False

    TARGET_VERSION = OpenBGPDLatestInstance.BGP_SPEAKER_VERSION
